from typing import Literal

import orjson


async def post_api_v5_trade_order(
    self,
    inst_id: str,
    td_mode: str,
    side: str,
    ord_type: str,
    sz: str,
    **kwargs,
):
    """
    Place a new order
    https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-place-order

    {'arg': {'channel': 'orders', 'instType': 'ANY', 'uid': '611800569950521616'}, 'data': [{'instType': 'SWAP', 'instId': 'BTC-USDT-SWAP', 'tgtCcy': '', 'ccy': '', 'ordId': '1993784914940116992', 'clOrdId': '', 'algoClOrdId': '', 'algoId': '', 'tag': '', 'px': '80000', 'sz': '0.1', 'notionalUsd': '80.0128', 'ordType': 'limit', 'side': 'buy', 'posSide': 'long', 'tdMode': 'cross', 'accFillSz': '0', 'fillNotionalUsd': '', 'avgPx': '0', 'state': 'canceled', 'lever': '3', 'pnl': '0', 'feeCcy': 'USDT', 'fee': '0', 'rebateCcy': 'USDT', 'rebate': '0', 'category': 'normal', 'uTime': '1731921825881', 'cTime': '1731921820806', 'source': '', 'reduceOnly': 'false', 'cancelSource': '1', 'quickMgnType': '', 'stpId': '', 'stpMode': 'cancel_maker', 'attachAlgoClOrdId': '', 'lastPx': '91880', 'isTpLimit': 'false', 'slTriggerPx': '', 'slTriggerPxType': '', 'tpOrdPx': '', 'tpTriggerPx': '', 'tpTriggerPxType': '', 'slOrdPx': '', 'fillPx': '', 'tradeId': '', 'fillSz': '0', 'fillTime': '', 'fillPnl': '0', 'fillFee': '0', 'fillFeeCcy': '', 'execType': '', 'fillPxVol': '', 'fillPxUsd': '', 'fillMarkVol': '', 'fillFwdPx': '', 'fillMarkPx': '', 'amendSource': '', 'reqId': '', 'amendResult': '', 'code': '0', 'msg': '', 'pxType': '', 'pxUsd': '', 'pxVol': '', 'linkedAlgoOrd': {'algoId': ''}, 'attachAlgoOrds': []}]}
    """
    endpoint = "/api/v5/trade/order"
    payload = {
        "instId": inst_id,
        "tdMode": td_mode,
        "side": side,
        "ordType": ord_type,
        "sz": sz,
        "tag": "f50cdd72d3b6BCDE",
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_cancel_order(
    self, inst_id: str, ord_id: str | None = None, cl_ord_id: str | None = None
):
    """
    Cancel an existing order
    https://www.okx.com/docs-v5/en/#order-book-trading-trade-post-cancel-order
    """
    endpoint = "/api/v5/trade/cancel-order"
    payload = {"instId": inst_id}
    if ord_id:
        payload["ordId"] = ord_id
    if cl_ord_id:
        payload["clOrdId"] = cl_ord_id

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_batch_orders(
    self,
    payload: list[dict[str, str]],
):
    """
    Place orders in batches. Maximum 20 orders can be placed per request.
    Request parameters should be passed in the form of an array. Orders will be placed in turn
    https://www.okx.com/docs-v5/en/?python#order-book-trading-trade-post-place-multiple-orders

    {'arg': {'channel': 'orders', 'instType': 'ANY', 'uid': '611800569950521616'}, 'data': [{'instType': 'SWAP', 'instId': 'BTC-USDT-SWAP', 'tgtCcy': '', 'ccy': '', 'ordId': '1993784914940116992', 'clOrdId': '', 'algoClOrdId': '', 'algoId': '', 'tag': '', 'px': '80000', 'sz': '0.1', 'notionalUsd': '80.0128', 'ordType': 'limit', 'side': 'buy', 'posSide': 'long', 'tdMode': 'cross', 'accFillSz': '0', 'fillNotionalUsd': '', 'avgPx': '0', 'state': 'canceled', 'lever': '3', 'pnl': '0', 'feeCcy': 'USDT', 'fee': '0', 'rebateCcy': 'USDT', 'rebate': '0', 'category': 'normal', 'uTime': '1731921825881', 'cTime': '1731921820806', 'source': '', 'reduceOnly': 'false', 'cancelSource': '1', 'quickMgnType': '', 'stpId': '', 'stpMode': 'cancel_maker', 'attachAlgoClOrdId': '', 'lastPx': '91880', 'isTpLimit': 'false', 'slTriggerPx': '', 'slTriggerPxType': '', 'tpOrdPx': '', 'tpTriggerPx': '', 'tpTriggerPxType': '', 'slOrdPx': '', 'fillPx': '', 'tradeId': '', 'fillSz': '0', 'fillTime': '', 'fillPnl': '0', 'fillFee': '0', 'fillFeeCcy': '', 'execType': '', 'fillPxVol': '', 'fillPxUsd': '', 'fillMarkVol': '', 'fillFwdPx': '', 'fillMarkPx': '', 'amendSource': '', 'reqId': '', 'amendResult': '', 'code': '0', 'msg': '', 'pxType': '', 'pxUsd': '', 'pxVol': '', 'linkedAlgoOrd': {'algoId': ''}, 'attachAlgoOrds': []}]}
    """
    endpoint = "/api/v5/trade/batch-orders"

    for order in payload:
        order["tag"] = "f50cdd72d3b6BCDE"

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_cancel_batch_orders(
    self,
    body: list[dict[str, str]],
):
    """
    Cancel multiple orders in batches
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-cancel-multiple-orders
    """
    endpoint = "/api/v5/trade/cancel-batch-orders"

    raw = await self._fetch("POST", endpoint, payload=body, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_amend_order(
    self,
    inst_id: str,
    **kwargs,
):
    """
    Modify an existing order
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-amend-order
    """
    endpoint = "/api/v5/trade/amend-order"
    payload = {
        "instId": inst_id,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_amend_batch_orders(
    self,
    body: list[dict[str, str]],
):
    """
    Modify multiple orders in batches
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-amend-multiple-orders
    """
    endpoint = "/api/v5/trade/amend-batch-orders"

    raw = await self._fetch("POST", endpoint, payload=body, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_close_position(
    self,
    inst_id: str,
    mgn_mode: Literal["cross", "isolated"],
    **kwargs,
):
    """
    Close the position of an instrument via a market order
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-close-positions
    """
    endpoint = "/api/v5/trade/close-position"
    payload = {
        "instId": inst_id,
        "mgnMode": mgn_mode,
        "tag": "f50cdd72d3b6BCDE",
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_order(
    self,
    inst_id: str,
    ord_id: str | None = None,
    cl_ord_id: str | None = None,
):
    """
    Retrieve order details.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-order-details
    """
    endpoint = "/api/v5/trade/order"
    payload = {"instId": inst_id}
    if ord_id:
        payload["ordId"] = ord_id
    if cl_ord_id:
        payload["clOrdId"] = cl_ord_id

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_orders_pending(
    self,
    **kwargs,
):
    """
    Retrieve all incomplete orders under the current account.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-order-list
    """
    endpoint = "/api/v5/trade/orders-pending"
    payload = {
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_orders_history(
    self,
    inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"],
    **kwargs,
):
    """
    Get completed orders which are placed in the last 7 days, including those placed 7 days ago but completed in the last 7 days.
    The incomplete orders that have been canceled are only reserved for 2 hours.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-order-history-last-7-days
    """
    endpoint = "/api/v5/trade/orders-history"
    payload = {
        "instType": inst_type,
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_orders_history_archive(
    self,
    inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"],
    **kwargs,
):
    """
    Get completed orders which are placed 7 days ago or earlier.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-order-history-last-3-months
    """
    endpoint = "/api/v5/trade/orders-history-archive"
    payload = {
        "instType": inst_type,
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_fills(
    self,
    **kwargs,
):
    """
    Retrieve recently-filled transaction details in the last 3 day.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-transaction-details-last-3-days
    """
    endpoint = "/api/v5/trade/fills"
    payload = {
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_fills_history(
    self,
    inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"],
    **kwargs,
):
    """
    Retrieve recently-filled transaction details in the last 3 months.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-transaction-details-last-3-months
    """
    endpoint = "/api/v5/trade/fills-history"
    payload = {
        "instType": inst_type,
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_easy_convert_currency_list(
    self,
    **kwargs,
):
    """
    Get list of small convertibles and mainstream currencies. Only applicable to the crypto balance less than $10.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-easy-convert-currency-list
    """
    endpoint = "/api/v5/trade/easy-convert-currency-list"
    payload = {
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_easy_convert(
    self,
    from_ccy: list[str],
    to_ccy: str,
    source: str = "1",
):
    """
    Convert small currencies to mainstream currencies.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-place-easy-convert
    """
    endpoint = "/api/v5/trade/easy-convert"
    payload = {
        "fromCcy": from_ccy,
        "toCcy": to_ccy,
        "source": source,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_easy_convert_history(
    self,
    **kwargs,
):
    """
    Get the history and status of easy convert trades in the past 7 days.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-easy-convert-history
    """
    endpoint = "/api/v5/trade/easy-convert-history"
    payload = {
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_one_click_repay_currency_list(
    self,
    **kwargs,
):
    """
    Get list of debt currency data and repay currencies. Debt currencies include both cross and isolated debts.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-one-click-repay-currency-list
    """
    endpoint = "/api/v5/trade/one-click-repay-currency-list"
    payload = {
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_one_click_repay(
    self,
    debt_ccy: list[str],
    repay_ccy: str,
):
    """
    Trade one-click repay to repay cross debts. Isolated debts are not applicable. The maximum repayment amount is based on the remaining available balance of funding and trading accounts.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-trade-one-click-repay
    """
    endpoint = "/api/v5/trade/one-click-repay"
    payload = {
        "debtCcy": debt_ccy,
        "repayCcy": repay_ccy,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_one_click_repay_history(
    self,
    **kwargs,
):
    """
    Get the history and status of one-click repay trades in the past 7 days.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-one-click-repay-history
    """
    endpoint = "/api/v5/trade/one-click-repay-history"
    payload = {
        **kwargs,
    }

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_mass_cancel(
    self,
    inst_type: str,
    inst_family: str,
    **kwargs,
):
    """
    Cancel all the MMP pending orders of an instrument family.
    Only applicable to Option in Portfolio Margin mode, and MMP privilege is required.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-mass-cancel-order
    """
    endpoint = "/api/v5/trade/mass-cancel"
    payload = {
        "instType": inst_type,
        "instFamily": inst_family,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_trade_cancel_all_after(
    self,
    time_out: str,
    **kwargs,
):
    """
    Cancel all pending orders after the countdown timeout. Applicable to all trading symbols through order book (except Spread trading)
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-cancel-all-after
    """
    endpoint = "/api/v5/trade/cancel-all-after"
    payload = {
        "timeOut": time_out,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_trade_account_rate_limit(self):
    """
    Get account rate limit related information.
    Only new order requests and amendment order requests will be counted towards this limit. For batch order requests consisting of multiple orders, each order will be counted individually.
    For details, please refer to Fill ratio based sub-account rate limit
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-get-account-rate-limit
    """
    endpoint = "/api/v5/trade/account-rate-limit"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)

    # POST /api/v5/trade/order-precheck


async def post_api_v5_trade_order_precheck(
    self,
    inst_id: str,
    td_mode: str,
    side: str,
    ord_type: str,
    sz: str,
    **kwargs,
):
    """
    This endpoint is used to precheck the account information before and after placing the order.
    Only applicable to Multi-currency margin mode, and Portfolio margin mode.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-trade-post-order-precheck
    """
    endpoint = "/api/v5/trade/order-precheck"
    payload = {
        "instId": inst_id,
        "tdMode": td_mode,
        "side": side,
        "ordType": ord_type,
        "sz": sz,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)



class BinanceError(Exception):
    """
    The base class for all Binance specific errors.
    """

    def __init__(self, status, message, headers):
        super().__init__(message)
        self.status = status
        self.message = message
        self.headers = headers


class BinanceServerError(BinanceError):
    """
    Represents an Binance specific 500 series HTTP error.
    """

    def __init__(self, status, message, headers):
        super().__init__(status, message, headers)


class BinanceClientError(BinanceError):
    """
    Represents an Binance specific 400 series HTTP error.
    """

    def __init__(self, status, message, headers):
        super().__init__(status, message, headers)



from typing import Dict, Any
import orjson
import hmac
import base64
import asyncio
import aiohttp
from urllib.parse import urljoin, urlencode, unquote

from nexusbridge.base import ApiClient
from nexusbridge.okx.constants import OkxUrl
from nexusbridge.okx.error import OkxHttpError, OkxRequestError


class OkxApiClient(ApiClient):
    def __init__(
        self,
        api_key: str = None,
        secret: str = None,
        passphrase: str = None,
        url: OkxUrl = OkxUrl.DEMO,
        timeout: int = 10,
        max_rate: int | None = None,
    ):
        super().__init__(
            api_key=api_key,
            secret=secret,
            timeout=timeout,
            max_rate=max_rate,
        )

        self._base_url = url.base_url
        self._passphrase = passphrase
        self._testnet = url.is_testnet
        self._headers = {
            "Content-Type": "application/json",
            "User-Agent": "TradingBot/1.0",
        }

    def _generate_signature(self, message: str) -> str:
        mac = hmac.new(
            bytes(self._secret, encoding="utf8"),
            bytes(message, encoding="utf-8"),
            digestmod="sha256",
        )
        return base64.b64encode(mac.digest()).decode()

    async def _get_signature(
        self, ts: str, method: str, request_path: str, payload: Dict[str, Any] = None
    ) -> str:
        body = ""
        if method == "POST":
            body = {}
        if payload:
            body = orjson.dumps(payload).decode()

        sign_str = f"{ts}{method}{request_path}{body}"
        signature = self._generate_signature(sign_str)
        return signature

    def _get_timestamp(self) -> str:
        return (
            self._clock.utc_now()
            .isoformat(timespec="milliseconds")
            .replace("+00:00", "Z")
        )

    async def _get_headers(
        self, ts: str, method: str, request_path: str, payload: Dict[str, Any] = None
    ) -> Dict[str, Any]:
        headers = self._headers
        signature = await self._get_signature(ts, method, request_path, payload)
        headers.update(
            {
                "OK-ACCESS-KEY": self._api_key,
                "OK-ACCESS-SIGN": signature,
                "OK-ACCESS-TIMESTAMP": ts,
                "OK-ACCESS-PASSPHRASE": self._passphrase,
            }
        )
        if self._testnet:
            headers["x-simulated-trading"] = "1"
        return headers

    async def _fetch(
        self,
        method: str,
        endpoint: str,
        payload: list[Dict[str, Any]] | Dict[str, Any] = None,
        signed: bool = False,
    ) -> bytes:
        if self._limiter:
            await self._limiter.wait()

        self._init_session()
        url = urljoin(self._base_url, endpoint)
        request_path = endpoint
        headers = self._headers
        timestamp = self._get_timestamp()

        payload = payload or {}

        payload_json = (
            unquote(urlencode(payload)) if method == "GET" else orjson.dumps(payload)
        )

        if method == "GET":
            url += f"?{payload_json}"
            request_path = (
                f"{request_path}?{payload_json}" if payload_json != "" else request_path
            )
            payload_json = None
            # when method is GET, payload should be empty
            payload = {}

        if signed and self._api_key:
            headers = await self._get_headers(timestamp, method, request_path, payload)

        try:
            self._log.debug(
                f"Request {method} Url: {url} Headers: {headers} Payload: {payload_json}"
            )

            response = await self._session.request(
                method=method,
                url=url,
                headers=headers,
                data=payload_json,
            )
            raw = await response.read()
            if response.status >= 400:
                raise OkxHttpError(
                    status_code=response.status,
                    message=orjson.loads(raw),
                    headers=response.headers,
                )
            okx_response = orjson.loads(raw)
            if okx_response["code"] == "0":
                return raw
            else:
                if len(okx_response["data"]) == 0:
                    raise OkxRequestError(
                        error_code=okx_response["code"],
                        status_code=okx_response["code"],
                        message=okx_response["msg"],
                    )
                else:
                    for data in okx_response["data"]:
                        raise OkxRequestError(
                            error_code=data["sCode"],
                            status_code=response.status,
                            message=data["sMsg"],
                        )
        except aiohttp.ClientError as e:
            self._log.error(f"Client Error {method} Url: {url} {e}")
            raise
        except asyncio.TimeoutError:
            self._log.error(f"Timeout {method} Url: {url}")
            raise
        except Exception as e:
            self._log.error(f"Error {method} Url: {url} {e}")
            raise

    # AlgoTradingApi
    from nexusbridge.okx.api.algo_trading import post_api_v5_trade_order_algo
    from nexusbridge.okx.api.algo_trading import post_api_v5_trade_cancel_algos
    from nexusbridge.okx.api.algo_trading import post_api_v5_trade_amend_algos
    from nexusbridge.okx.api.algo_trading import post_api_v5_trade_cancel_advance_algos
    from nexusbridge.okx.api.algo_trading import get_api_v5_trade_order_algo
    from nexusbridge.okx.api.algo_trading import get_api_v5_trade_orders_algo_pending
    from nexusbridge.okx.api.algo_trading import get_api_v5_trade_orders_algo_history
    from nexusbridge.okx.api.algo_trading import post_api_v5_trading_bot_grid_order_algo

    # CopyTradingApi
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_current_subpositions,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_subpositions_history,
    )
    from nexusbridge.okx.api.copy_trading import post_api_v5_copytrading_algo_order
    from nexusbridge.okx.api.copy_trading import (
        post_api_v5_copytrading_close_subposition,
    )
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_instruments
    from nexusbridge.okx.api.copy_trading import post_api_v5_copytrading_set_instruments
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_profit_sharing_details,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_total_profit_sharing,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_unrealized_profit_sharing_details,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_total_unrealized_profit_sharing,
    )
    from nexusbridge.okx.api.copy_trading import (
        post_api_v5_copytrading_apply_lead_trading,
    )
    from nexusbridge.okx.api.copy_trading import (
        post_api_v5_copytrading_stop_lead_trading,
    )
    from nexusbridge.okx.api.copy_trading import (
        post_api_v5_copytrading_amend_profit_sharing_ratio,
    )
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_config
    from nexusbridge.okx.api.copy_trading import (
        post_api_v5_copytrading_first_copy_settings,
    )
    from nexusbridge.okx.api.copy_trading import (
        post_api_v5_copytrading_amend_copy_settings,
    )
    from nexusbridge.okx.api.copy_trading import (
        post_api_v5_copytrading_stop_copy_trading,
    )
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_copy_settings
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_current_lead_traders,
    )
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_public_config
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_public_lead_traders,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_public_weekly_pnl,
    )
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_public_pnl
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_public_stats
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_public_preference_currency,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_public_current_subpositions,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_public_subpositions_history,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_public_copy_traders,
    )
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_lead_traders
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_weekly_pnl
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_pnl
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_stats
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_preference_currency,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_performance_current_subpositions,
    )
    from nexusbridge.okx.api.copy_trading import (
        get_api_v5_copytrading_performance_subpositions_history,
    )
    from nexusbridge.okx.api.copy_trading import get_api_v5_copytrading_copy_traders

    # GridTradingApi
    from nexusbridge.okx.api.grid_trading import post_api_v5_trading_bot_grid_order_algo
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_amend_order_algo,
    )
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_stop_order_algo,
    )
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_close_position,
    )
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_cancel_close_order,
    )
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_order_instant_trigger,
    )
    from nexusbridge.okx.api.grid_trading import (
        get_api_v5_trading_bot_grid_orders_algo_pending,
    )
    from nexusbridge.okx.api.grid_trading import (
        get_api_v5_trading_bot_grid_orders_algo_history,
    )
    from nexusbridge.okx.api.grid_trading import (
        get_api_v5_trading_bot_grid_orders_algo_details,
    )
    from nexusbridge.okx.api.grid_trading import get_api_v5_trading_bot_grid_sub_orders
    from nexusbridge.okx.api.grid_trading import get_api_v5_trading_bot_grid_positions
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_withdraw_income,
    )
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_compute_margin_balance,
    )
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_margin_balance,
    )
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_adjust_investment,
    )
    from nexusbridge.okx.api.grid_trading import get_api_v5_trading_bot_grid_ai_param
    from nexusbridge.okx.api.grid_trading import (
        post_api_v5_trading_bot_grid_min_investment,
    )
    from nexusbridge.okx.api.grid_trading import get_api_v5_trading_bot_rsi_back_testing
    from nexusbridge.okx.api.grid_trading import get_api_v5_trading_bot_grid_quantity

    # MarketDataApi
    from nexusbridge.okx.api.market_data import get_api_v5_market_tickers
    from nexusbridge.okx.api.market_data import get_api_v5_market_ticker
    from nexusbridge.okx.api.market_data import get_api_v5_market_books
    from nexusbridge.okx.api.market_data import get_api_v5_market_books_full
    from nexusbridge.okx.api.market_data import get_api_v5_market_candles
    from nexusbridge.okx.api.market_data import get_api_v5_market_history_candles
    from nexusbridge.okx.api.market_data import get_api_v5_market_trades
    from nexusbridge.okx.api.market_data import get_api_v5_market_history_trades
    from nexusbridge.okx.api.market_data import (
        get_api_v5_market_option_instrument_family_trades,
    )
    from nexusbridge.okx.api.market_data import get_api_v5_public_option_trades
    from nexusbridge.okx.api.market_data import get_api_v5_market_platform_24_volume
    from nexusbridge.okx.api.market_data import get_api_v5_market_call_auction_details

    # PublicDataApi
    from nexusbridge.okx.api.public_data import get_api_v5_public_instruments
    from nexusbridge.okx.api.public_data import (
        get_api_v5_public_delivery_exercise_history,
    )
    from nexusbridge.okx.api.public_data import get_api_v5_public_open_interest
    from nexusbridge.okx.api.public_data import get_api_v5_public_funding_rate
    from nexusbridge.okx.api.public_data import get_api_v5_public_funding_rate_history
    from nexusbridge.okx.api.public_data import get_api_v5_public_price_limit
    from nexusbridge.okx.api.public_data import get_api_v5_public_opt_summary
    from nexusbridge.okx.api.public_data import get_api_v5_public_estimated_price
    from nexusbridge.okx.api.public_data import (
        get_api_v5_public_discount_rate_interest_free_quota,
    )
    from nexusbridge.okx.api.public_data import get_api_v5_public_time
    from nexusbridge.okx.api.public_data import get_api_v5_public_mark_price
    from nexusbridge.okx.api.public_data import get_api_v5_public_position_tiers
    from nexusbridge.okx.api.public_data import (
        get_api_v5_public_interest_rate_loan_quota,
    )
    from nexusbridge.okx.api.public_data import get_api_v5_public_underlying
    from nexusbridge.okx.api.public_data import get_api_v5_public_insurance_fund
    from nexusbridge.okx.api.public_data import get_api_v5_public_convert_contract_coin
    from nexusbridge.okx.api.public_data import get_api_v5_public_instrument_tick_bands
    from nexusbridge.okx.api.public_data import get_api_v5_public_premium_history
    from nexusbridge.okx.api.public_data import get_api_v5_market_index_tickers
    from nexusbridge.okx.api.public_data import get_api_v5_market_index_candles
    from nexusbridge.okx.api.public_data import get_api_v5_market_history_index_candles
    from nexusbridge.okx.api.public_data import get_api_v5_market_mark_price_candles
    from nexusbridge.okx.api.public_data import (
        get_api_v5_market_history_mark_price_candles,
    )
    from nexusbridge.okx.api.public_data import get_api_v5_market_open_oracle
    from nexusbridge.okx.api.public_data import get_api_v5_market_exchange_rate
    from nexusbridge.okx.api.public_data import get_api_v5_market_index_components
    from nexusbridge.okx.api.public_data import get_api_v5_public_economic_calendar

    # RecurringBuyApi
    from nexusbridge.okx.api.recurring_buy import (
        post_api_v5_trading_bot_signal_create_signal,
    )
    from nexusbridge.okx.api.recurring_buy import (
        post_api_v5_trading_bot_recurring_order_algo,
    )
    from nexusbridge.okx.api.recurring_buy import (
        post_api_v5_trading_bot_recurring_amend_order_algo,
    )
    from nexusbridge.okx.api.recurring_buy import (
        post_api_v5_trading_bot_recurring_stop_order_algo,
    )
    from nexusbridge.okx.api.recurring_buy import (
        get_api_v5_trading_bot_recurring_orders_algo_pending,
    )
    from nexusbridge.okx.api.recurring_buy import (
        get_api_v5_trading_bot_recurring_orders_algo_history,
    )
    from nexusbridge.okx.api.recurring_buy import (
        get_api_v5_trading_bot_recurring_orders_algo_details,
    )
    from nexusbridge.okx.api.recurring_buy import (
        get_api_v5_trading_bot_recurring_sub_orders,
    )

    # SignalBotTradingApi
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_create_signal,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        get_api_v5_trading_bot_signal_signals,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_order_algo,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_stop_order_algo,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_margin_balance,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_amend_tp_sl,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_set_instruments,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        get_api_v5_trading_bot_signal_orders_algo_details,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        get_api_v5_trading_bot_signal_orders_algo_pending,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        get_api_v5_trading_bot_signal_orders_algo_history,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        get_api_v5_trading_bot_signal_positions,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        get_api_v5_trading_bot_signal_positions_history,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_close_position,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_sub_order,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        post_api_v5_trading_bot_signal_cancel_sub_order,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        get_api_v5_trading_bot_signal_sub_orders,
    )
    from nexusbridge.okx.api.signal_bot_trading import (
        get_api_v5_trading_bot_signal_event_history,
    )

    # TradeApi
    from nexusbridge.okx.api.trade import post_api_v5_trade_order
    from nexusbridge.okx.api.trade import post_api_v5_trade_cancel_order
    from nexusbridge.okx.api.trade import post_api_v5_trade_batch_orders
    from nexusbridge.okx.api.trade import post_api_v5_trade_cancel_batch_orders
    from nexusbridge.okx.api.trade import post_api_v5_trade_amend_order
    from nexusbridge.okx.api.trade import post_api_v5_trade_amend_batch_orders
    from nexusbridge.okx.api.trade import post_api_v5_trade_close_position
    from nexusbridge.okx.api.trade import get_api_v5_trade_order
    from nexusbridge.okx.api.trade import get_api_v5_trade_orders_pending
    from nexusbridge.okx.api.trade import get_api_v5_trade_orders_history
    from nexusbridge.okx.api.trade import get_api_v5_trade_orders_history_archive
    from nexusbridge.okx.api.trade import get_api_v5_trade_fills
    from nexusbridge.okx.api.trade import get_api_v5_trade_fills_history
    from nexusbridge.okx.api.trade import get_api_v5_trade_easy_convert_currency_list
    from nexusbridge.okx.api.trade import post_api_v5_trade_easy_convert
    from nexusbridge.okx.api.trade import get_api_v5_trade_easy_convert_history
    from nexusbridge.okx.api.trade import get_api_v5_trade_one_click_repay_currency_list
    from nexusbridge.okx.api.trade import post_api_v5_trade_one_click_repay
    from nexusbridge.okx.api.trade import get_api_v5_trade_one_click_repay_history
    from nexusbridge.okx.api.trade import post_api_v5_trade_mass_cancel
    from nexusbridge.okx.api.trade import post_api_v5_trade_cancel_all_after
    from nexusbridge.okx.api.trade import get_api_v5_trade_account_rate_limit
    from nexusbridge.okx.api.trade import post_api_v5_trade_order_precheck

    # TradingAccountApi
    from nexusbridge.okx.api.trading_account import get_api_v5_account_instruments
    from nexusbridge.okx.api.trading_account import get_api_v5_account_balance
    from nexusbridge.okx.api.trading_account import get_api_v5_account_positions
    from nexusbridge.okx.api.trading_account import get_api_v5_account_positions_history
    from nexusbridge.okx.api.trading_account import get_api_v5_account_position_risk
    from nexusbridge.okx.api.trading_account import get_api_v5_account_bills
    from nexusbridge.okx.api.trading_account import get_api_v5_account_bills_archive
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_bills_history_archive,
    )
    from nexusbridge.okx.api.trading_account import (
        get_api_v5_account_bills_history_archive,
    )
    from nexusbridge.okx.api.trading_account import get_api_v5_account_config
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_set_position_mode,
    )
    from nexusbridge.okx.api.trading_account import post_api_v5_account_set_leverage
    from nexusbridge.okx.api.trading_account import get_api_v5_account_max_size
    from nexusbridge.okx.api.trading_account import get_api_v5_account_max_avail_size
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_position_margin_balance,
    )
    from nexusbridge.okx.api.trading_account import get_api_v5_account_leverage_info
    from nexusbridge.okx.api.trading_account import (
        get_api_v5_account_adjust_leverage_info,
    )
    from nexusbridge.okx.api.trading_account import get_api_v5_account_max_loan
    from nexusbridge.okx.api.trading_account import get_api_v5_account_trade_fee
    from nexusbridge.okx.api.trading_account import get_api_v5_account_interest_accrued
    from nexusbridge.okx.api.trading_account import get_api_v5_account_interest_rate
    from nexusbridge.okx.api.trading_account import post_api_v5_account_set_greeks
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_set_isolated_mode,
    )
    from nexusbridge.okx.api.trading_account import get_api_v5_account_max_withdrawal
    from nexusbridge.okx.api.trading_account import get_api_v5_account_risk_state
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_quick_margin_borrow_repay,
    )
    from nexusbridge.okx.api.trading_account import (
        get_api_v5_account_quick_margin_borrow_repay_history,
    )
    from nexusbridge.okx.api.trading_account import get_api_v5_account_interest_limits
    from nexusbridge.okx.api.trading_account import (
        get_api_v5_account_fixed_loan_borrowing_limit,
    )
    from nexusbridge.okx.api.trading_account import (
        get_api_v5_account_fixed_loan_borrowing_quote,
    )
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_fixed_loan_borrowing_order,
    )
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_fixed_loan_amend_borrowing_order,
    )
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_fixed_loan_manual_reborrow,
    )
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_fixed_loan_repay_borrowing_order,
    )
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_fixed_loan_convert_to_market_loan,
    )
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_fixed_loan_reduce_liabilities,
    )
    from nexusbridge.okx.api.trading_account import (
        get_api_v5_account_fixed_loan_borrowing_orders_list,
    )
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_spot_manual_borrow_repay,
    )
    from nexusbridge.okx.api.trading_account import post_api_v5_account_set_auto_repay
    from nexusbridge.okx.api.trading_account import (
        get_api_v5_account_spot_borrow_repay_history,
    )
    from nexusbridge.okx.api.trading_account import post_api_v5_account_position_builder
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_set_risk_offset_amt,
    )
    from nexusbridge.okx.api.trading_account import get_api_v5_account_greeks
    from nexusbridge.okx.api.trading_account import get_api_v5_account_position_tiers
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_set_risk_offset_type,
    )
    from nexusbridge.okx.api.trading_account import post_api_v5_account_activate_option
    from nexusbridge.okx.api.trading_account import post_api_v5_account_set_auto_loan
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_account_level_switch_preset,
    )
    from nexusbridge.okx.api.trading_account import (
        get_api_v5_account_set_account_switch_precheck,
    )
    from nexusbridge.okx.api.trading_account import (
        post_api_v5_account_set_account_level,
    )
    from nexusbridge.okx.api.trading_account import post_api_v5_account_mmp_reset
    from nexusbridge.okx.api.trading_account import post_api_v5_account_mmp_config
    from nexusbridge.okx.api.trading_account import get_api_v5_account_mmp_config

    # FinancialProductApi
    from nexusbridge.okx.api.financial_product import (
        get_api_v5_finance_savings_lending_rate_history,
    )
    from nexusbridge.okx.api.financial_product import (
        get_api_v5_finance_savings_lending_rate_summary,
    )



from typing import Literal

from nexusbridge.binance.constants import Side, OrderType


async def post_api_v3_order(self, symbol: str, side: Side, _type: OrderType, **kwargs):
    """
    Send in a new order.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints
    """
    endpoint = "/api/v3/order"
    payload = {"symbol": symbol, "side": side.value, "type": _type.value, **kwargs}

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_api_v3_order_test(
    self,
    symbol: str,
    side: Side,
    _type: OrderType,
    compute_commission_rates: bool = False,
    **kwargs,
):
    """
    Test new order creation and signature/recvWindow long. Creates and validates a new order but does not send it into the matching engine.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#test-new-order-trade
    """
    endpoint = "/api/v3/order/test"
    payload = {
        "computeCommissionRates": compute_commission_rates,
        "symbol": symbol,
        "side": side.value,
        "type": _type.value,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_api_v3_order(self, symbol: str, **kwargs):
    """
    Check an order's status.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#query-order-user_data
    """
    endpoint = "/api/v3/order"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def delete_api_v3_order(self, symbol: str, **kwargs):
    """
    Cancel an active order.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#cancel-order-trade
    """
    endpoint = "/api/v3/order"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("DELETE", endpoint, payload=payload, signed=True)
    return raw


async def delete_api_v3_open_orders(self, symbol: str, **kwargs):
    """
    Cancels all active orders on a symbol. This includes orders that are part of an order list.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#cancel-all-open-orders-on-a-symbol-trade
    """
    endpoint = "/api/v3/openOrders"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_api_v3_order_cancel_replace(
    self,
    symbol: str,
    side: Side,
    _type: OrderType,
    cancel_replace_mode: Literal["STOP_ON_FAILURE", "ALLOW_FAILURE"],
    **kwargs,
):
    """
    Cancels an existing order and places a new order on the same symbol.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#cancel-an-existing-order-and-send-a-new-order-trade
    """
    endpoint = "/api/v3/order/cancelReplace"
    payload = {
        "symbol": symbol,
        "side": side.value,
        "type": _type.value,
        "cancelReplaceMode": cancel_replace_mode,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def get_api_v3_open_orders(self, symbol: str, **kwargs):
    """
    Get all open orders on a symbol. Careful when accessing this with no symbol.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#current-open-orders-user_data
    """
    endpoint = "/api/v3/openOrders"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_api_v3_all_orders(self, symbol: str, **kwargs):
    """
    Get all account orders; active, canceled, or filled.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#all-orders-user_data
    """
    endpoint = "/api/v3/allOrders"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def post_api_v3_order_oco(
    self,
    symbol: str,
    side: Side,
    quantity: float,
    price: float,
    stop_price: float,
    **kwargs,
):
    """
    Post an OCO order
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#order-lists
    """
    endpoint = "/api/v3/order/oco"
    payload = {
        "symbol": symbol,
        "side": side.value,
        "quantity": quantity,
        "price": price,
        "stopPrice": stop_price,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw


async def post_api_v3_sor_order(
    self, symbol: str, side: Side, _type: OrderType, quantity: float, **kwargs
):
    """
    Places an order using smart order routing (SOR).
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/trading-endpoints#sor
    """
    endpoint = "/api/v3/sor/order"
    payload = {
        "symbol": symbol,
        "side": side.value,
        "type": _type.value,
        "quantity": quantity,
        **kwargs,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return raw



from typing import Literal

import orjson


async def get_api_v5_account_instruments(self, inst_type: str, **kwargs):
    """
    Retrieve available instruments info of current account.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-instruments
    """
    endpoint = "/api/v5/account/instruments"
    payload = {"instType": inst_type, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_balance(self, ccy: str | None = None):
    """
    Retrieve a list of assets (with non-zero balance), remaining balance, and available amount in the trading account.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-balance
    """
    endpoint = "/api/v5/account/balance"
    payload = {}
    if ccy:
        payload["ccy"] = ccy
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_positions(self, **kwargs):
    """
    Retrieve information on your positions. When the account is in net mode, net positions will be displayed, and when the account is in long/short mode, long or short positions will be displayed. Return in reverse chronological order using ctime.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-positions
    """
    endpoint = "/api/v5/account/positions"
    payload = {**kwargs}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_positions_history(self, **kwargs):
    """
    Retrieve the updated position data for the last 3 months. Return in reverse chronological order using utime. Getting positions history is supported under Portfolio margin mode since 04:00 AM (UTC) on November 11, 2024
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-positions-history
    """
    endpoint = "/api/v5/account/positions-history"
    payload = {**kwargs}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_position_risk(
    self, inst_type: Literal["MARGIN", "SWAP", "FUTURES", "OPTION"] | None = None
):
    """
    Retrieve the risk limit of the account.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-account-and-position-risk
    """
    endpoint = "/api/v5/account/account-position-risk"
    payload = {}
    if inst_type:
        payload["instType"] = inst_type

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_bills(self, **kwargs):
    """
    Retrieve the bills of the account. The bill refers to all transaction records that result in changing the balance of an account. Pagination is supported, and the response is sorted with the most recent first. This endpoint can retrieve data from the last 7 days.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-bills-details-last-7-days
    """
    endpoint = "/api/v5/account/bills"
    payload = {**kwargs}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_bills_archive(self, **kwargs):
    """
    Retrieve the account’s bills. The bill refers to all transaction records that result in changing the balance of an account. Pagination is supported, and the response is sorted with most recent first. This endpoint can retrieve data from the last 3 months.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-bills-details-last-3-months
    """
    endpoint = "/api/v5/account/bills-archive"
    payload = {**kwargs}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_bills_history_archive(
    self,
    year: str,
    quarter: Literal["Q1", "Q2", "Q3", "Q4"],
):
    """
    Apply for bill data since 1 February, 2021 except for the current quarter.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-apply-bills-details-since-2021
    """
    endpoint = "/api/v5/account/bills-history-archive"
    payload = {"year": year, "quarter": quarter}

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_bills_history_archive(
    self,
    year: str,
    quarter: Literal["Q1", "Q2", "Q3", "Q4"],
):
    """
    Apply for bill data since 1 February, 2021 except for the current quarter.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-bills-details-since-2021
    """
    endpoint = "/api/v5/account/bills-history-archive"
    payload = {"year": year, "quarter": quarter}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_config(self):
    """
    Retrieve current account configuration.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-account-configuration
    """
    endpoint = "/api/v5/account/config"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_position_mode(
    self, pos_mode: Literal["long_short_mode", "net_mode"]
):
    """
    Spot and futures mode and Multi-currency mode: FUTURES and SWAP support both long/short mode and net mode. In net mode, users can only have positions in one direction; In long/short mode, users can hold positions in long and short directions.
    Portfolio margin mode: FUTURES and SWAP only support net mode
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-position-mode
    """
    endpoint = "/api/v5/account/set-position-mode"
    payload = {"posMode": pos_mode}

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_leverage(
    self, mgn_mode: Literal["isolated", "cross"], lever: str, ccy: str, **kwargs
):
    """
    There are 10 different scenarios for leverage setting:
    1. Set leverage for MARGIN instruments under isolated-margin trade mode at pairs level.
    2. Set leverage for MARGIN instruments under cross-margin trade mode and Spot mode (enabled borrow) at currency level.
    3. Set leverage for MARGIN instruments under cross-margin trade mode and Spot and futures mode account mode at pairs level.
    4. Set leverage for MARGIN instruments under cross-margin trade mode and Multi-currency margin at currency level.
    5. Set leverage for MARGIN instruments under cross-margin trade mode and Portfolio margin at currency level.
    6. Set leverage for FUTURES instruments under cross-margin trade mode at underlying level.
    7. Set leverage for FUTURES instruments under isolated-margin trade mode and buy/sell position mode at contract level.
    8. Set leverage for FUTURES instruments under isolated-margin trade mode and long/short position mode at contract and position side level.
    9. Set leverage for SWAP instruments under cross-margin trade at contract level.
    10. Set leverage for SWAP instruments under isolated-margin trade mode and buy/sell position mode at contract level.
    11. Set leverage for SWAP instruments under isolated-margin trade mode and long/short position mode at contract and position side level.
    Note that the request parameter posSide is only required when margin mode is isolated in long/short position mode for FUTURES/SWAP instruments (see scenario 8 and 11 above).
    Please refer to the request examples on the right for each case.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-leverage
    """
    endpoint = "/api/v5/account/set-leverage"
    payload = {"lever": lever, "mgnMode": mgn_mode, "ccy": ccy, **kwargs}

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_max_size(
    self,
    inst_id: str,
    td_mode: Literal["cross", "isolated", "cash", "spot_isolated"],
    **kwargs,
):
    """
    The maximum quantity to buy or sell. It corresponds to the "sz" from placement.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-maximum-order-quantity
    """
    endpoint = "/api/v5/account/max-size"
    payload = {"instId": inst_id, "tdMode": td_mode, **kwargs}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_max_avail_size(
    self,
    inst_id: str,
    td_mode: Literal["cross", "isolated", "cash", "spot_isolated"],
    **kwargs,
):
    """
    Available balance for isolated margin positions and SPOT, available equity for cross margin positions.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-maximum-available-balance-equity
    """
    endpoint = "/api/v5/account/max-avail-size"
    payload = {"instId": inst_id, "tdMode": td_mode, **kwargs}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_position_margin_balance(
    self,
    inst_id: str,
    pos_side: Literal["long", "short"],
    _type: Literal["add", "reduce"],
    amt: str,
    ccy: str | None = None,
):
    """
    Increase or decrease the margin of the isolated position. Margin reduction may result in the change of the actual leverage.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-increase-decrease-margin
    """
    endpoint = "/api/v5/account/position/margin-balance"
    payload = {
        "instId": inst_id,
        "posSide": pos_side,
        "type": _type,
        "amt": amt,
    }
    if ccy:
        payload["ccy"] = ccy

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_leverage_info(
    self, mgn_mode: Literal["isolated", "cross"], **kwargs
):
    """
    Get leverage
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-leverage
    """
    endpoint = "/api/v5/account/leverage-info"
    payload = {"mgnMode": mgn_mode, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_adjust_leverage_info(
    self,
    inst_type: Literal["MARGIN", "SWAP", "FUTURES"],
    mgn_mode: Literal["isolated", "cross"],
    lever: str,
    **kwargs,
):
    """
    Get leverage estimated info
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-leverage-estimated-info
    """
    endpoint = "/api/v5/account/adjust-leverage-info"
    payload = {"instType": inst_type, "mgnMode": mgn_mode, "lever": lever, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_max_loan(
    self, mgn_mode: Literal["isolated", "cross"], **kwargs
):
    """
    Get the maximum loan of instrument
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-the-maximum-loan-of-instrument
    """
    endpoint = "/api/v5/account/max-loan"
    payload = {"mgnMode": mgn_mode, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_trade_fee(
    self,
    inst_type: Literal["SPOT", "MARGIN", "SWAP", "FUTURES", "OPTION"],
    rule_type: Literal["normal", "pre_market"] | None = None,
    **kwargs,
):
    """
    Get fee rates
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-fee-rates
    """
    endpoint = "/api/v5/account/trade-fee"
    payload = {"instType": inst_type, **kwargs}
    if rule_type:
        payload["ruleType"] = rule_type
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_interest_accrued(self, **kwargs):
    """
    Get interest accrued data. Only data within the last one year can be obtained.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-interest-accrued-data
    """
    endpoint = "/api/v5/account/interest-accrued"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_interest_rate(self, ccy: str | None = None):
    """
    Get the user's current leveraged currency borrowing market interest rate
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-interest-rate
    """
    endpoint = "/api/v5/account/interest-rate"
    payload = {}
    if ccy:
        payload["ccy"] = ccy
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_greeks(
    self,
    greeks_type: Literal["PA", "BS"],
):
    """
    Set the display type of Greeks.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-greeks-pa-bs
    """
    endpoint = "/api/v5/account/set-greeks"
    payload = {"greeksType": greeks_type}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_isolated_mode(
    self, _type: Literal["MARGIN", "CONTRACTS"], iso_mode: str = "automatic"
):
    """
    Set isolated margin mode
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-isolated-margin-mode
    """
    endpoint = "/api/v5/account/set-isolated-mode"
    payload = {"type": _type, "isoMode": iso_mode}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_max_withdrawal(self, ccy: str | None = None):
    """
    Retrieve the maximum transferable amount from trading account to funding account. If no currency is specified, the transferable amount of all owned currencies will be returned.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-maximum-withdrawals
    """
    endpoint = "/api/v5/account/max-withdrawal"
    payload = {}
    if ccy:
        payload["ccy"] = ccy
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_risk_state(self):
    """
    Only applicable to Portfolio margin account
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-account-risk-state
    """
    endpoint = "/api/v5/account/risk-state"
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_quick_margin_borrow_repay(
    self,
    inst_id: str,
    ccy: str,
    amt: str,
    side: Literal["borrow", "repay"],
):
    """
    Please note that this endpoint will be deprecated soon.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-manual-borrow-and-repay-in-quick-margin-mode
    """
    endpoint = "/api/v5/account/quick-margin-borrow-repay"
    payload = {
        "instId": inst_id,
        "ccy": ccy,
        "amt": amt,
        "side": side,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_quick_margin_borrow_repay_history(self, **kwargs):
    """
    Get borrow and repay history in Quick Margin Mode
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-borrow-and-repay-history-in-quick-margin-mode
    """
    endpoint = "/api/v5/account/quick-margin-borrow-repay-history"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_interest_limits(
    self, _type: str | None = None, ccy: str | None = None
):
    """
    Get borrow interest and limit
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-borrow-interest-and-limit
    """
    endpoint = "/api/v5/account/interest-limits"
    payload = {}
    if _type:
        payload["type"] = _type
    if ccy:
        payload["ccy"] = ccy
    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_fixed_loan_borrowing_limit(self):
    """
    Get fixed loan borrow limit
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-fixed-loan-borrow-limit
    """
    endpoint = "/api/v5/account/fixed-loan/borrowing-limit"

    raw = await self._fetch("GET", endpoint, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_fixed_loan_borrowing_quote(
    self, _type: Literal["normal", "reborrow"], **kwargs
):
    """
    Get fixed loan borrow quote
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-fixed-loan-borrow-quote
    """
    endpoint = "/api/v5/account/fixed-loan/borrowing-quote"
    payload = {"type": _type, **kwargs}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_fixed_loan_borrowing_order(
    self, ccy: str, amt: str, max_rate: str, term: str = "30D", **kwargs
):
    """
    For new borrowing orders, they belong to the IOC (immediately close and cancel the remaining) type. For renewal orders, they belong to the FOK (Fill-or-kill) type.
    Order book may refer to Get lending volume (public).
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-place-fixed-loan-borrowing-order
    """
    endpoint = "/api/v5/account/fixed-loan/borrowing-order"
    payload = {"ccy": ccy, "amt": amt, "term": term, "maxRate": max_rate, **kwargs}

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_fixed_loan_amend_borrowing_order(
    self, ord_id: str, **kwargs
):
    """
    Amend fixed loan borrowing order
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-amend-fixed-loan-borrowing-order
    """
    endpoint = "/api/v5/account/fixed-loan/amend-borrowing-order"
    payload = {"ordId": ord_id, **kwargs}

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_fixed_loan_manual_reborrow(
    self, ord_id: str, max_rate: str
):
    """
    Manual renew fixed loan borrowing order
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-manual-renew-fixed-loan-borrowing-order
    """
    endpoint = "/api/v5/account/fixed-loan/manual-reborrow"
    payload = {"ordId": ord_id, "maxRate": max_rate}

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_fixed_loan_repay_borrowing_order(
    self,
    ord_id: str,
):
    """
    Repay fixed loan borrowing order
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-repay-fixed-loan-borrowing-order
    """
    endpoint = "/api/v5/account/fixed-loan/repay-borrowing-order"
    payload = {
        "ordId": ord_id,
    }

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_fixed_loan_convert_to_market_loan(
    self,
    ord_id: str,
):
    """
    Convert fixed loan to market loan
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-convert-fixed-loan-to-market-loan
    """
    endpoint = "/api/v5/account/fixed-loan/convert-to-market-loan"
    payload = {
        "ordId": ord_id,
    }

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_fixed_loan_reduce_liabilities(
    self,
    ord_id: str,
    pending_repay: bool,
):
    """
    Provide the function of "setting pending repay state / canceling pending repay state" for fixed loan order.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-reduce-liabilities-for-fixed-loan
    """
    endpoint = "/api/v5/account/fixed-loan/reduce-liabilities"
    payload = {
        "ordId": ord_id,
        "pendingRepay": pending_repay,
    }

    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_fixed_loan_borrowing_orders_list(self, **kwargs):
    """
    Get fixed loan borrow order list
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-fixed-loan-borrow-order-list
    """
    endpoint = "/api/v5/account/fixed-loan/borrowing-orders-list"
    payload = {**kwargs}

    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_spot_manual_borrow_repay(
    self,
    ccy: str,
    amt: str,
    side: Literal["borrow", "repay"],
):
    """
    Only applicable to Spot mode (enabled borrowing)
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-manual-borrow-repay
    """
    endpoint = "/api/v5/account/spot-manual-borrow-repay"
    payload = {
        "ccy": ccy,
        "amt": amt,
        "side": side,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_auto_repay(
    self,
    auto_repay: bool,
):
    """
    Only applicable to Spot mode (enabled borrowing)
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-auto-repay
    """
    endpoint = "/api/v5/account/set-auto-repay"
    payload = {
        "autoRepay": auto_repay,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_spot_borrow_repay_history(self, **kwargs):
    """
    Retrieve the borrow/repay history under Spot mode
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-borrow-repay-history
    """
    endpoint = "/api/v5/account/spot-borrow-repay-history"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_position_builder(self, **kwargs):
    """
    Calculates portfolio margin information for virtual position/assets or current position of the user.
    You can add up to 200 virtual positions and 200 virtual assets in one request.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-position-builder-new
    """
    endpoint = "/api/v5/account/position-builder"
    payload = {**kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_risk_offset_amt(
    self,
    cl_spot_in_use_amt: str,
    ccy: str,
):
    """
    Set risk offset amount. This does not represent the actual spot risk offset amount. Only applicable to Portfolio Margin Mode.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-risk-offset-amount
    """
    endpoint = "/api/v5/account/set-riskOffset-amt"
    payload = {"clSpotInUseAmt": cl_spot_in_use_amt, "ccy": ccy}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_greeks(self, ccy: str | None = None):
    """
    Retrieve a greeks list of all assets in the account.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-greeks
    """
    endpoint = "/api/v5/account/greeks"
    payload = {}
    if ccy:
        payload["ccy"] = ccy
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_position_tiers(
    self, inst_type: Literal["SWAP", "FUTURES", "OPTION"], **kwargs
):
    """
    Retrieve cross position limitation of SWAP/FUTURES/OPTION under Portfolio margin mode.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-pm-position-limitation
    """
    endpoint = "/api/v5/account/position-tiers"
    payload = {"instType": inst_type, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_risk_offset_type(
    self, _type: Literal["1", "2", "3", "4"]
):
    """
    Configure the risk offset type in portfolio margin mode.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-risk-offset-type
    """
    endpoint = "/api/v5/account/set-riskOffset-type"
    payload = {"type": _type}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_activate_option(self):
    """
    Activate option
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-activate-option
    """
    endpoint = "/api/v5/account/activate-option"

    raw = await self._fetch("POST", endpoint, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_auto_loan(
    self,
    auto_loan: bool | None = None,
):
    """
    Only applicable to Multi-currency margin and Portfolio margin
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-auto-loan
    """
    endpoint = "/api/v5/account/set-auto-loan"
    payload = {}
    if auto_loan is not None:
        payload["autoLoan"] = auto_loan
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_account_level_switch_preset(
    self, acct_lv: Literal["2", "3", "4"], **kwargs
):
    """
    Pre-set the required information for account mode switching. When switching from Spot and futures mode / Multi-currency margin mode to Portfolio margin mode, you may optionally pre-set riskOffetType. When switching from Portfolio margin mode back to Spot and futures mode / Multi-currency margin mode, and if there are existing cross-margin contract positions, it is mandatory to pre-set leverage.
    If the user does not follow the required settings, they will receive an error message during the pre-check or when setting the account mode.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-preset-account-mode-switch
    """
    endpoint = "/api/v5/account/account-level-switch-preset"
    payload = {"acctLv": acct_lv, **kwargs}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_set_account_switch_precheck(
    self,
    acct_lv: Literal["1", "2", "3", "4"],
):
    """
    Retrieve precheck information for account mode switching.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-preset-account-mode-switch
    """
    endpoint = "/api/v5/account/set-account-switch-precheck"
    payload = {
        "acctLv": acct_lv,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_set_account_level(
    self,
    acct_lv: Literal["1", "2", "3", "4"],
):
    """
    You need to set on the Web/App for the first set of every account mode. If users plan to switch account modes while holding positions, they should first call the preset endpoint to conduct necessary settings, then call the precheck endpoint to get unmatched information, margin check, and other related information, and finally call the account mode switch endpoint to switch account modes.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-account-mode
    """
    endpoint = "/api/v5/account/set-account-level"
    payload = {
        "acctLv": acct_lv,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_mmp_reset(
    self, inst_family: str, inst_type: str = "OPTION"
):
    """
    You can unfreeze by this endpoint once MMP is triggered.
    Only applicable to Option in Portfolio Margin mode, and MMP privilege is required.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-reset-mmp-status
    """
    endpoint = "/api/v5/account/mmp-reset"
    payload = {"instFamily": inst_family, "instType": inst_type}
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def post_api_v5_account_mmp_config(
    self,
    inst_family: str,
    time_interval: str,
    frozen_interval: str,
    qty_limit: str,
):
    """
    This endpoint is used to set MMP configure
    Only applicable to Option in Portfolio Margin mode, and MMP privilege is required.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-set-mmp
    """
    endpoint = "/api/v5/account/mmp-config"
    payload = {
        "instFamily": inst_family,
        "timeInterval": time_interval,
        "frozenInterval": frozen_interval,
        "qtyLimit": qty_limit,
    }
    raw = await self._fetch("POST", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)


async def get_api_v5_account_mmp_config(self, inst_family: str | None = None):
    """
    This endpoint is used to get MMP configure information
    Only applicable to Option in Portfolio Margin mode, and MMP privilege is required.
    https://www.okx.com/docs-v5/en/?shell#trading-account-rest-api-get-mmp-config
    """
    endpoint = "/api/v5/account/mmp-config"
    payload = {}
    if inst_family:
        payload["instFamily"] = inst_family
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return orjson.loads(raw)



from typing import Literal
from urllib.parse import urlencode

from nexusbridge.binance.constants import OrderType


async def depth(self, symbol: str, _id: int | None = None, **kwargs):
    """
    Get current order book. Note that this request returns limited market depth
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/websocket-api
    """
    payload = {"symbol": symbol, **kwargs}
    return await self.send("depth", params=payload, _id=_id)


async def ticker_price(self, _id: int | None = None, **kwargs):
    """
    Latest price for a symbol or symbols.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/websocket-api/Symbol-Price-Ticker
    """
    payload = {**kwargs}
    return await self.send("ticker.price", params=payload, _id=_id)


async def ticker_book(self, _id: int | None = None, **kwargs):
    """
    Best price/qty on the order book for a symbol or symbols.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/market-data/websocket-api/Symbol-Order-Book-Ticker
    """
    payload = {**kwargs}
    return await self.send("ticker.book", params=payload, _id=_id)


async def order_place(
    self,
    symbol: str,
    side: Literal["BUY", "SELL"],
    _type: OrderType,
    _id: int | None = None,
    **kwargs,
):
    """
    Send in a new order.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/websocket-api
    """
    payload = {
        "apiKey": self.key,
        "symbol": symbol,
        "side": side,
        "type": _type.value,
        "timestamp": self._clock.timestamp_ms(),
        **kwargs,
    }
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.place", params=payload, _id=_id)


async def order_modify(
    self,
    symbol: str,
    side: Literal["BUY", "SELL"],
    _type: OrderType,
    quantity: float,
    price: float,
    _id: int | None = None,
    **kwargs,
):
    """
    Order modify function, currently only LIMIT order modification is supported, modified orders will be reordered in the match queue
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/websocket-api/Modify-Order
    """
    payload = {
        "apiKey": self.key,
        "symbol": symbol,
        "side": side,
        "type": _type.value,
        "quantity": quantity,
        "price": price,
        "timestamp": self._clock.timestamp_ms(),
        **kwargs,
    }
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.modify", params=payload, _id=_id)


async def order_cancel(self, symbol: str, _id: int | None = None, **kwargs):
    """
    Cancel an active order.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/websocket-api/Cancel-Order
    """
    payload = {
        "apiKey": self.key,
        "symbol": symbol,
        "timestamp": self._clock.timestamp_ms(),
        **kwargs,
    }
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.cancel", params=payload, _id=_id)


async def order_status(self, symbol: str, _id: int | None = None, **kwargs):
    """
    Check an order's status.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/websocket-api/Query-Order
    """
    payload = {
        "apiKey": self.key,
        "symbol": symbol,
        "timestamp": self._clock.timestamp_ms(),
        **kwargs,
    }
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("order.status", params=payload, _id=_id)


async def account_position_v2(self, _id: int | None = None, **kwargs):
    """
    Get current position information(only symbol that has position or open orders will be returned).
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/websocket-api/Position-Info-V2
    """
    payload = {"apiKey": self.key, "timestamp": self._clock.timestamp_ms(), **kwargs}
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("v2/account.position", params=payload, _id=_id)


async def account_position(self, _id: int | None = None, **kwargs):
    """
    Get current position information.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/trade/websocket-api/Position-Information
    """
    payload = {"apiKey": self.key, "timestamp": self._clock.timestamp_ms(), **kwargs}
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("account.position", params=payload, _id=_id)


async def user_data_stream_start(
    self,
    _id: int | None = None,
):
    """
    Start a new user data stream. The stream will close after 60 minutes unless a keepalive is sent. If the account has an active listenKey, that listenKey will be returned and its validity will be extended for 60 minutes.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/user-data-streams/Start-User-Data-Stream-Wsp
    """
    payload = {
        "apiKey": self.key,
    }
    return await self.send("userDataStream.start", params=payload, _id=_id)


async def user_data_stream_ping(
    self,
    listen_key: str,
    _id: int | None = None,
):
    """
    Keepalive a user data stream to prevent a timeout. User data streams will close after 60 minutes. It's recommended to send a ping about every 30 minutes.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/user-data-streams/Ping-Pong
    """
    payload = {
        "listenKey": listen_key,
        "apiKey": self.key,
    }
    return await self.send("userDataStream.ping", params=payload, _id=_id)


async def user_data_stream_stop(
    self,
    listen_key: str,
    _id: int | None = None,
):
    """
    Close out a user data stream.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/user-data-streams/Close-User-Data-Stream-Wsp
    """
    payload = {
        "listenKey": listen_key,
        "apiKey": self.key,
    }
    return await self.send("userDataStream.stop", params=payload, _id=_id)


async def account_balance_v2(self, _id: int | None = None, **kwargs):
    """
    Query account balance info
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/websocket-api
    """
    payload = {"apiKey": self.key, "timestamp": self._clock.timestamp_ms(), **kwargs}
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("account.balance", params=payload, _id=_id)


async def account_balance(self, _id: int | None = None, **kwargs):
    """
    Query account balance info
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/websocket-api/Futures-Account-Balance
    """
    payload = {"apiKey": self.key, "timestamp": self._clock.timestamp_ms(), **kwargs}
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("account.balance", params=payload, _id=_id)


async def account_status_v2(self, _id: int | None = None, **kwargs):
    """
    Get current account information. User in single-asset/ multi-assets mode will see different value, see comments in response section for detail.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/websocket-api/Account-Information-V2
    """
    payload = {"apiKey": self.key, "timestamp": self._clock.timestamp_ms(), **kwargs}
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("account.status", params=payload, _id=_id)


async def account_status(self, _id: int | None = None, **kwargs):
    """
    Get current account information. User in single-asset/ multi-assets mode will see different value, see comments in response section for detail.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/account/websocket-api/Account-Information
    """
    payload = {"apiKey": self.key, "timestamp": self._clock.timestamp_ms(), **kwargs}
    encoded_payload = urlencode(payload)
    payload["signature"] = self._get_sign(encoded_payload)
    return await self.send("account.status", params=payload, _id=_id)


async def agg_trade(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    The Aggregate Trade Streams push market trade information that is aggregated for fills with same price and taking side every 100 milliseconds. Only market trades will be aggregated, which means the insurance fund trades and ADL trades won't be aggregated.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Aggregate-Trade-Streams
    """
    stream_name = "{}@aggTrade".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mark_price(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Mark price and funding rate for a single symbol pushed every 3 seconds or every second.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Mark-Price-Stream
    """
    stream_name = "{}@markPrice".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mark_price_arr(self, subscription: bool = True, _id: int | None = None):
    """
    Mark price and funding rate for all symbols pushed every 3 seconds or every second.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Mark-Price-Stream-for-All-market
    """
    stream_name = "!markPrice@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def kline(
    self,
    symbol: str,
    interval: Literal[
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    The Kline/Candlestick Stream push updates to the current klines/candlestick every 250 milliseconds (if existing).
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Kline-Candlestick-Streams
    """
    stream_name = "{}@kline_{}".format(symbol.lower(), interval)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def continuous_kline(
    self,
    pair: str,
    contract_type: Literal["perpetual", "current_quarter", "next_quarter"],
    interval: Literal[
        "1m",
        "3m",
        "5m",
        "15m",
        "30m",
        "1h",
        "2h",
        "4h",
        "6h",
        "8h",
        "12h",
        "1d",
        "3d",
        "1w",
        "1M",
    ],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Continuous Contract Kline/Candlestick Streams
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Continuous-Contract-Kline-Candlestick-Streams
    """
    stream_name = "{}_{}@continuousKline_{}".format(
        pair.lower(), contract_type, interval
    )
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mini_ticker(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    24hr rolling window mini-ticker statistics for a single symbol. These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Individual-Symbol-Mini-Ticker-Stream
    """
    stream_name = "{}@miniTicker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ticker_arr(self, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window ticker statistics for all symbols. These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before. Note that only tickers that have changed will be present in the array.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/All-Market-Tickers-Streams
    """
    stream_name = "!ticker@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def ticker(self, symbol: str, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window ticker statistics for a single symbol. These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Individual-Symbol-Ticker-Streams
    """
    stream_name = "{}@ticker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def mini_ticker_arr(self, subscription: bool = True, _id: int | None = None):
    """
    24hr rolling window mini-ticker statistics for all symbols. These are NOT the statistics of the UTC day, but a 24hr rolling window from requestTime to 24hrs before. Note that only tickers that have changed will be present in the array.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/All-Market-Mini-Tickers-Stream
    """
    stream_name = "!miniTicker@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def book_ticker(
    self, symbol: str | list[str], subscription: bool = True, _id: int | None = None
):
    """
    Pushes any update to the best bid or ask's price or quantity in real-time for a specified symbol.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Individual-Symbol-Book-Ticker-Streams
    """
    if isinstance(symbol, list):
        stream_name = ["{}@bookTicker".format(s.lower()) for s in symbol]
    if isinstance(symbol, str):
        stream_name = "{}@bookTicker".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def book_ticker_arr(self, subscription: bool = True, _id: int | None = None):
    """
    Pushes any update to the best bid or ask's price or quantity in real-time for all symbols.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/All-Book-Tickers-Stream
    """
    stream_name = "!bookTicker"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def force_orders(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    The Liquidation Order Snapshot Streams push force liquidation order information for specific symbol. For each symbol，only the latest one liquidation order within 1000ms will be pushed as the snapshot. If no liquidation happens in the interval of 1000ms, no stream will be pushed.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Liquidation-Order-Streams
    """
    stream_name = "{}@forceOrder".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def force_orders_arr(self, subscription: bool = True, _id: int | None = None):
    """
    The All Liquidation Order Snapshot Streams push force liquidation order information for all symbols in the market. For each symbol，only the latest one liquidation order within 1000ms will be pushed as the snapshot. If no liquidation happens in the interval of 1000ms, no stream will be pushed.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/All-Market-Liquidation-Order-Streams
    """
    stream_name = "!forceOrder@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def depth_levels(
    self,
    symbol: str,
    levels: Literal[5, 10, 20],
    subscription: bool = True,
    _id: int | None = None,
):
    """
    Top <levels> bids and asks, Valid <levels> are 5, 10, or 20.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Partial-Book-Depth-Streams
    """
    stream_name = "{}@depth{}".format(symbol.lower(), levels)
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def diff_depth(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    ids and asks, pushed every 250 milliseconds, 500 milliseconds, 100 milliseconds (if existing)
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Diff-Book-Depth-Streams
    """
    stream_name = "{}@depth".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def composite_index(
    self, symbol: str, subscription: bool = True, _id: int | None = None
):
    """
    Composite index information for index symbols pushed every second.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Composite-Index-Symbol-Information-Streams
    """
    stream_name = "{}@compositeIndex".format(symbol.lower())
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def contract_info(self, subscription: bool = True, _id: int | None = None):
    """
    ContractInfo stream pushes when contract info updates(listing/settlement/contract bracket update). bks field only shows up when bracket gets updated.
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Contract-Info-Stream
    """
    stream_name = "!contractInfo"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)


async def asset_index(self, subscription: bool = True, _id: int | None = None):
    """
    Asset index for multi-assets mode user
    https://developers.binance.com/docs/derivatives/usds-margined-futures/websocket-market-streams/Multi-Assets-Mode-Asset-Index
    """
    stream_name = "!assetIndex@arr"
    if subscription:
        await self._subscribe(stream_name, _id)
    else:
        await self._unsubscribe(stream_name, _id)



import base64
import hmac
import hashlib
from base64 import b64encode
from Crypto.PublicKey import RSA
from Crypto.Hash import SHA256
from Crypto.Signature import pkcs1_15
from cryptography.hazmat.primitives.serialization import load_pem_private_key


def hmac_hashing(secret, payload):
    m = hmac.new(secret.encode("utf-8"), payload.encode("utf-8"), hashlib.sha256)
    return m.hexdigest()


def rsa_signature(private_key, payload, private_key_pass=None):
    private_key = RSA.import_key(private_key, passphrase=private_key_pass)
    h = SHA256.new(payload.encode("utf-8"))
    signature = pkcs1_15.new(private_key).sign(h)
    return b64encode(signature).decode("utf-8")


def ed_25519(private_key: str, payload, private_key_pass: str = None):
    key = load_pem_private_key(
        private_key.encode("utf-8"), password=private_key_pass.encode("utf-8")
    )
    return base64.b64encode(key.sign(payload.encode("ASCII"))).decode("utf-8")



import orjson


async def get_api_v5_market_tickers(self, inst_type: str, **kwargs):
    """
    Retrieve the latest price snapshot, best bid/ask price, and trading volume in the last 24 hours.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-tickers
    """
    endpoint = "/api/v5/market/tickers"
    payload = {"instType": inst_type, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_ticker(self, inst_id: str):
    """
    Retrieve the latest price snapshot, best bid/ask price, and trading volume in the last 24 hours.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-ticker
    """
    endpoint = "/api/v5/market/ticker"
    payload = {"instId": inst_id}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_books(self, inst_id: str, sz: str | None = None):
    """
    Retrieve order book of the instrument.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-order-book
    """
    endpoint = "/api/v5/market/books"
    payload = {
        "instId": inst_id,
    }
    if sz:
        payload["sz"] = sz
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_books_full(self, inst_id: str, sz: str | None = None):
    """
    Retrieve order book of the instrument. The data will be updated once a second.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-full-order-book
    """
    endpoint = "/api/v5/market/books-full"
    payload = {
        "instId": inst_id,
    }
    if sz:
        payload["sz"] = sz
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_candles(self, inst_id: str, **kwargs):
    """
    Retrieve the candlestick charts. This endpoint can retrieve the latest 1,440 data entries. Charts are returned in groups based on the requested bar.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-candlesticks
    """
    endpoint = "/api/v5/market/candles"
    payload = {"instId": inst_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_history_candles(self, inst_id: str, **kwargs):
    """
    Retrieve history candlestick charts from recent years(It is last 3 months supported for 1s candlestick).
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-candlesticks-history
    """
    endpoint = "/api/v5/market/history-candles"
    payload = {"instId": inst_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_trades(
    self,
    inst_id: str,
    limit: str | None = None,
):
    """
    Retrieve the recent transactions of an instrument.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-trades
    """
    endpoint = "/api/v5/market/trades"
    payload = {
        "instId": inst_id,
    }
    if limit:
        payload["limit"] = limit
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_history_trades(self, inst_id: str, **kwargs):
    """
    Retrieve the recent transactions of an instrument from the last 3 months with pagination.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-trades-history
    """
    endpoint = "/api/v5/market/history-trades"
    payload = {"instId": inst_id, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_option_instrument_family_trades(self, inst_family: str):
    """
    Retrieve the recent transactions of an instrument under same instFamily. The maximum is 100.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-option-trades-by-instrument-family
    """
    endpoint = "/api/v5/market/option/instrument-family-trades"
    payload = {"instFamily": inst_family}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_public_option_trades(self, **kwargs):
    """
    The maximum is 100.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-option-trades
    """
    endpoint = "/api/v5/public/option-trades"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_platform_24_volume(self):
    """
    The 24-hour trading volume is calculated on a rolling basis.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-24h-total-volume
    """
    endpoint = "/api/v5/market/platform-24-volume"
    raw = await self._fetch("GET", endpoint, signed=False)
    return orjson.loads(raw)


async def get_api_v5_market_call_auction_details(self, inst_id: str):
    """
    Retrieve call auction details.
    https://www.okx.com/docs-v5/en/?shell#order-book-trading-market-data-get-call-auction-details
    """
    endpoint = "/api/v5/market/call-auction-details"
    payload = {"instId": inst_id}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=False)
    return orjson.loads(raw)



from typing import Literal, Callable, Any

from nexusbridge.binance.base import BinanceApiClient, BinanceWSClient
from nexusbridge.binance.constants import (
    BinanceUrl,
    ALL_URL,
    BinanceInstrumentType,
    KeyType,
)


class UmTradingApi(BinanceApiClient):
    def __init__(
        self,
        key=None,
        secret=None,
        binance_url: Literal[BinanceUrl.MAIN, BinanceUrl.TEST] = BinanceUrl.MAIN,
        **kwargs,
    ):
        if "base_url" not in kwargs:
            kwargs["base_url"] = ALL_URL[BinanceInstrumentType.Derivatives_UM].get_url(
                binance_url
            )
        super().__init__(key=key, secret=secret, **kwargs)

    # General endpoints
    from nexusbridge.binance.um.general import get_fapi_v1_exchangeInfo

    # Market Data endpoints
    from nexusbridge.binance.um.market_data import get_fapi_v1_ping
    from nexusbridge.binance.um.market_data import get_fapi_v1_time
    from nexusbridge.binance.um.market_data import get_fapi_v1_exchange_info
    from nexusbridge.binance.um.market_data import get_fapi_v1_depth
    from nexusbridge.binance.um.market_data import get_fapi_v1_trades
    from nexusbridge.binance.um.market_data import get_fapi_v1_historical_trades
    from nexusbridge.binance.um.market_data import get_fapi_v1_agg_trades
    from nexusbridge.binance.um.market_data import get_fapi_v1_klines
    from nexusbridge.binance.um.market_data import get_fapi_v1_continuous_klines
    from nexusbridge.binance.um.market_data import get_fapi_v1_index_price_klines
    from nexusbridge.binance.um.market_data import get_fapi_v1_mark_price_klines
    from nexusbridge.binance.um.market_data import get_fapi_v1_premium_index_klines
    from nexusbridge.binance.um.market_data import get_fapi_v1_premium_index
    from nexusbridge.binance.um.market_data import get_fapi_v1_funding_rate
    from nexusbridge.binance.um.market_data import get_fapi_v1_funding_info
    from nexusbridge.binance.um.market_data import get_fapi_v1_ticker_24hr
    from nexusbridge.binance.um.market_data import get_fapi_v1_ticker_price
    from nexusbridge.binance.um.market_data import get_fapi_v2_ticker_price
    from nexusbridge.binance.um.market_data import get_fapi_v1_book_ticker
    from nexusbridge.binance.um.market_data import get_fapi_v1_delivery_price
    from nexusbridge.binance.um.market_data import get_fapi_v1_open_interest
    from nexusbridge.binance.um.market_data import get_fapi_v1_open_interest_hist
    from nexusbridge.binance.um.market_data import (
        get_futures_data_top_long_short_position_ratio,
    )
    from nexusbridge.binance.um.market_data import (
        get_futures_data_top_long_short_account_ratio,
    )
    from nexusbridge.binance.um.market_data import (
        get_futures_data_global_long_short_account_ratio,
    )
    from nexusbridge.binance.um.market_data import (
        get_futures_data_taker_long_short_ratio,
    )
    from nexusbridge.binance.um.market_data import get_fapi_v1_lvt_klines
    from nexusbridge.binance.um.market_data import get_fapi_v1_index_info
    from nexusbridge.binance.um.market_data import get_fapi_v1_asset_index
    from nexusbridge.binance.um.market_data import get_fapi_v1_constituents

    # Trading endpoints
    from nexusbridge.binance.um.trade import post_fapi_v1_order
    from nexusbridge.binance.um.trade import post_fapi_v1_batch_order
    from nexusbridge.binance.um.trade import put_fapi_v1_order
    from nexusbridge.binance.um.trade import put_fapi_v1_batch_order
    from nexusbridge.binance.um.trade import get_fapi_v1_order_amendment
    from nexusbridge.binance.um.trade import delete_fapi_v1_order
    from nexusbridge.binance.um.trade import delete_fapi_v1_batch_order
    from nexusbridge.binance.um.trade import delete_fapi_v1_all_open_orders
    from nexusbridge.binance.um.trade import post_fapi_v1_countdown_cancel_all
    from nexusbridge.binance.um.trade import get_fapi_v1_order
    from nexusbridge.binance.um.trade import get_fapi_v1_all_orders
    from nexusbridge.binance.um.trade import get_fapi_v1_open_orders
    from nexusbridge.binance.um.trade import get_fapi_v1_open_order
    from nexusbridge.binance.um.trade import get_fapi_v1_force_orders
    from nexusbridge.binance.um.trade import get_fapi_v1_user_trades
    from nexusbridge.binance.um.trade import post_fapi_v1_margin_type
    from nexusbridge.binance.um.trade import post_fapi_v1_position_side_dual
    from nexusbridge.binance.um.trade import post_fapi_v1_leverage
    from nexusbridge.binance.um.trade import post_fapi_v1_multi_assets_margin
    from nexusbridge.binance.um.trade import post_fapi_v1_position_margin
    from nexusbridge.binance.um.trade import get_fapi_v2_position_risk
    from nexusbridge.binance.um.trade import get_fapi_v3_position_risk
    from nexusbridge.binance.um.trade import get_fapi_v1_adl_quantile
    from nexusbridge.binance.um.trade import get_fapi_v1_position_margin_history
    from nexusbridge.binance.um.trade import post_fapi_v1_order_test

    # User Data Streams endpoints
    from nexusbridge.binance.um.user_data_stream import post_fapi_v1_listen_key
    from nexusbridge.binance.um.user_data_stream import put_fapi_v1_listen_key
    from nexusbridge.binance.um.user_data_stream import delete_fapi_v1_listen_key

    # Account endpoints
    from nexusbridge.binance.um.account import get_fapi_v3_balance
    from nexusbridge.binance.um.account import get_fapi_v2_balance
    from nexusbridge.binance.um.account import get_fapi_v3_account
    from nexusbridge.binance.um.account import get_fapi_v2_account
    from nexusbridge.binance.um.account import get_fapi_v1_commission_rate
    from nexusbridge.binance.um.account import get_fapi_v1_account_config
    from nexusbridge.binance.um.account import get_fapi_v1_symbol_config
    from nexusbridge.binance.um.account import get_fapi_v1_rate_limit_order
    from nexusbridge.binance.um.account import get_fapi_v1_leverage_bracket
    from nexusbridge.binance.um.account import get_fapi_v1_multi_assets_margin
    from nexusbridge.binance.um.account import get_fapi_v1_position_side_dual
    from nexusbridge.binance.um.account import get_fapi_v1_income
    from nexusbridge.binance.um.account import get_fapi_v1_api_trading_status
    from nexusbridge.binance.um.account import get_fapi_v1_income_asyn
    from nexusbridge.binance.um.account import get_fapi_v1_income_asyn_id
    from nexusbridge.binance.um.account import get_fapi_v1_order_asyn
    from nexusbridge.binance.um.account import get_fapi_v1_order_asyn_id
    from nexusbridge.binance.um.account import get_fapi_v1_trade_asyn
    from nexusbridge.binance.um.account import get_fapi_v1_trade_asyn_id
    from nexusbridge.binance.um.account import post_fapi_v1_fee_burn
    from nexusbridge.binance.um.account import get_fapi_v1_fee_burn
    from nexusbridge.binance.um.account import get_fapi_v1_pm_account_info

    # Convert endpoints
    from nexusbridge.binance.um.convert import get_fapi_v1_convert_exchange_info
    from nexusbridge.binance.um.convert import post_fapi_v1_convert_get_quote
    from nexusbridge.binance.um.convert import post_fapi_v1_convert_accept_quote
    from nexusbridge.binance.um.convert import get_fapi_v1_convert_order_status


class UmTradingWebsocket(BinanceWSClient):
    def __init__(
        self,
        handler: Callable[..., Any],
        key=None,
        secret=None,
        listen_key=None,
        private_key=None,
        private_key_passphrase=None,
        key_type: KeyType = KeyType.HMAC,
        binance_url: Literal[
            BinanceUrl.WS, BinanceUrl.TEST_WS, BinanceUrl.WS_STREAM
        ] = BinanceUrl.WS_STREAM,
        **kwargs,
    ):
        if "base_url" not in kwargs:
            self.baseUrl = ALL_URL[BinanceInstrumentType.Derivatives_UM].get_url(
                binance_url
            )
        else:
            self.baseUrl = kwargs["base_url"]
            kwargs.pop("base_url", None)
        if listen_key:
            self.baseUrl = f"{self.baseUrl}/{listen_key}"
        self.key = key
        self.secret = secret
        self.private_key = private_key
        self.private_key_passphrase = private_key_passphrase
        self.key_type = key_type
        super().__init__(
            url=self.baseUrl,
            key=key,
            key_type=self.key_type,
            handler=handler,
            private_key=self.private_key,
            private_key_passphrase=self.private_key_passphrase,
            secret=secret,
            **kwargs,
        )

    # wss://ws-fapi.binance.com/ws-fapi/v1
    from nexusbridge.binance.um.web_socket_stream import depth
    from nexusbridge.binance.um.web_socket_stream import ticker_price
    from nexusbridge.binance.um.web_socket_stream import ticker_book
    from nexusbridge.binance.um.web_socket_stream import order_place
    from nexusbridge.binance.um.web_socket_stream import order_modify
    from nexusbridge.binance.um.web_socket_stream import order_cancel
    from nexusbridge.binance.um.web_socket_stream import order_status
    from nexusbridge.binance.um.web_socket_stream import account_position_v2
    from nexusbridge.binance.um.web_socket_stream import account_position
    from nexusbridge.binance.um.web_socket_stream import user_data_stream_start
    from nexusbridge.binance.um.web_socket_stream import user_data_stream_ping
    from nexusbridge.binance.um.web_socket_stream import user_data_stream_stop
    from nexusbridge.binance.um.web_socket_stream import account_balance_v2
    from nexusbridge.binance.um.web_socket_stream import account_balance
    from nexusbridge.binance.um.web_socket_stream import account_status_v2
    from nexusbridge.binance.um.web_socket_stream import account_status

    # wss://fstream.binance.com/ws
    from nexusbridge.binance.um.web_socket_stream import agg_trade
    from nexusbridge.binance.um.web_socket_stream import mark_price
    from nexusbridge.binance.um.web_socket_stream import mark_price_arr
    from nexusbridge.binance.um.web_socket_stream import kline
    from nexusbridge.binance.um.web_socket_stream import continuous_kline
    from nexusbridge.binance.um.web_socket_stream import mini_ticker
    from nexusbridge.binance.um.web_socket_stream import ticker_arr
    from nexusbridge.binance.um.web_socket_stream import ticker
    from nexusbridge.binance.um.web_socket_stream import mini_ticker_arr
    from nexusbridge.binance.um.web_socket_stream import book_ticker
    from nexusbridge.binance.um.web_socket_stream import book_ticker_arr
    from nexusbridge.binance.um.web_socket_stream import force_orders
    from nexusbridge.binance.um.web_socket_stream import force_orders_arr
    from nexusbridge.binance.um.web_socket_stream import depth_levels
    from nexusbridge.binance.um.web_socket_stream import diff_depth
    from nexusbridge.binance.um.web_socket_stream import composite_index
    from nexusbridge.binance.um.web_socket_stream import contract_info
    from nexusbridge.binance.um.web_socket_stream import asset_index



from typing import Literal, Callable, Any

from nexusbridge.binance.base import BinanceApiClient, BinanceWSClient
from nexusbridge.binance.constants import (
    BinanceUrl,
    ALL_URL,
    BinanceInstrumentType,
    KeyType,
)


class SpotTradingApi(BinanceApiClient):
    def __init__(
        self,
        key=None,
        secret=None,
        binance_url: Literal[BinanceUrl.MAIN, BinanceUrl.TEST] = BinanceUrl.MAIN,
        **kwargs,
    ):
        if "base_url" not in kwargs:
            kwargs["base_url"] = ALL_URL[BinanceInstrumentType.SPOT].get_url(
                binance_url
            )
        super().__init__(key=key, secret=secret, **kwargs)

    # General endpoints
    from nexusbridge.binance.spot.general import get_api_v3_ping
    from nexusbridge.binance.spot.general import get_api_v3_time
    from nexusbridge.binance.spot.general import get_api_v3_exchangeInfo

    # Market Data endpoints
    from nexusbridge.binance.spot.market_data import get_api_v3_depth
    from nexusbridge.binance.spot.market_data import get_api_v3_trades
    from nexusbridge.binance.spot.market_data import get_api_v3_historical_trades
    from nexusbridge.binance.spot.market_data import get_api_v3_klines
    from nexusbridge.binance.spot.market_data import get_api_v3_ui_klines
    from nexusbridge.binance.spot.market_data import get_api_v3_avgPrice
    from nexusbridge.binance.spot.market_data import get_api_v3_ticker_24hr
    from nexusbridge.binance.spot.market_data import get_api_v3_ticker_tradingDay
    from nexusbridge.binance.spot.market_data import get_api_v3_ticker_price
    from nexusbridge.binance.spot.market_data import get_api_v3_ticker_bookTicker
    from nexusbridge.binance.spot.market_data import get_api_v3_ticker

    # Trading endpoints
    from nexusbridge.binance.spot.trading import post_api_v3_order
    from nexusbridge.binance.spot.trading import post_api_v3_order_test
    from nexusbridge.binance.spot.trading import get_api_v3_order
    from nexusbridge.binance.spot.trading import delete_api_v3_order
    from nexusbridge.binance.spot.trading import delete_api_v3_open_orders
    from nexusbridge.binance.spot.trading import post_api_v3_order_cancel_replace
    from nexusbridge.binance.spot.trading import get_api_v3_open_orders
    from nexusbridge.binance.spot.trading import get_api_v3_all_orders
    from nexusbridge.binance.spot.trading import post_api_v3_order_oco
    from nexusbridge.binance.spot.trading import post_api_v3_sor_order

    # Account Endpoints
    from nexusbridge.binance.spot.account import get_api_v3_account
    from nexusbridge.binance.spot.account import get_api_v3_my_trades
    from nexusbridge.binance.spot.account import get_api_v3_rate_limit_order
    from nexusbridge.binance.spot.account import get_api_v3_my_prevented_matches
    from nexusbridge.binance.spot.account import get_api_v3_my_allocations
    from nexusbridge.binance.spot.account import get_api_v3_account_commission

    # User data stream endpoints
    from nexusbridge.binance.spot.user_data_stream import post_api_v3_user_data_stream
    from nexusbridge.binance.spot.user_data_stream import put_api_v3_user_data_stream
    from nexusbridge.binance.spot.user_data_stream import delete_api_v3_user_data_stream


class BinanceSpotWSClient(BinanceWSClient):
    def __init__(
        self,
        handler: Callable[..., Any],
        key=None,
        secret=None,
        listen_key=None,
        private_key=None,
        private_key_passphrase=None,
        key_type: KeyType = KeyType.HMAC,
        binance_url: Literal[BinanceUrl.WS, BinanceUrl.TEST_WS] = BinanceUrl.WS,
        **kwargs,
    ):
        if "base_url" not in kwargs:
            self.baseUrl = ALL_URL[BinanceInstrumentType.SPOT].get_url(binance_url)
        else:
            self.baseUrl = kwargs["base_url"]
            kwargs.pop("base_url", None)
        if listen_key:
            self.baseUrl = f"{self.baseUrl}/{listen_key}"
        self.key = key
        self.secret = secret
        self.private_key = private_key
        self.private_key_passphrase = private_key_passphrase
        self.key_type = key_type
        super().__init__(
            url=self.baseUrl,
            key=key,
            key_type=self.key_type,
            handler=handler,
            private_key=self.private_key,
            private_key_passphrase=self.private_key_passphrase,
            secret=secret,
            **kwargs,
        )

    # wss://stream.binance.com:443
    from nexusbridge.binance.spot.web_socket_stream import agg_trade
    from nexusbridge.binance.spot.web_socket_stream import trade
    from nexusbridge.binance.spot.web_socket_stream import kline
    from nexusbridge.binance.spot.web_socket_stream import ui_kline
    from nexusbridge.binance.spot.web_socket_stream import mini_ticker
    from nexusbridge.binance.spot.web_socket_stream import mini_ticker_arr
    from nexusbridge.binance.spot.web_socket_stream import ticker
    from nexusbridge.binance.spot.web_socket_stream import ticker_arr
    from nexusbridge.binance.spot.web_socket_stream import ticker_window
    from nexusbridge.binance.spot.web_socket_stream import ticker_window_arr
    from nexusbridge.binance.spot.web_socket_stream import book_ticker
    from nexusbridge.binance.spot.web_socket_stream import avg_price
    from nexusbridge.binance.spot.web_socket_stream import depth
    from nexusbridge.binance.spot.web_socket_stream import diff_depth

    # wss://ws-api.binance.com:443/ws-api/v3
    from nexusbridge.binance.spot.web_socket_stream import order_book_api
    from nexusbridge.binance.spot.web_socket_stream import recent_trades_api
    from nexusbridge.binance.spot.web_socket_stream import historical_trades_api
    from nexusbridge.binance.spot.web_socket_stream import agg_trades_api
    from nexusbridge.binance.spot.web_socket_stream import klines_api
    from nexusbridge.binance.spot.web_socket_stream import ui_klines_api
    from nexusbridge.binance.spot.web_socket_stream import avg_price_api
    from nexusbridge.binance.spot.web_socket_stream import ticker_24hr_api
    from nexusbridge.binance.spot.web_socket_stream import ticker_trading_day_api
    from nexusbridge.binance.spot.web_socket_stream import ticker_api
    from nexusbridge.binance.spot.web_socket_stream import ticker_price_api
    from nexusbridge.binance.spot.web_socket_stream import ticker_book_api
    from nexusbridge.binance.spot.web_socket_stream import session_logon_api
    from nexusbridge.binance.spot.web_socket_stream import session_status_api
    from nexusbridge.binance.spot.web_socket_stream import session_logout_api
    from nexusbridge.binance.spot.web_socket_stream import order_place_api
    from nexusbridge.binance.spot.web_socket_stream import order_test_api
    from nexusbridge.binance.spot.web_socket_stream import order_status_api
    from nexusbridge.binance.spot.web_socket_stream import order_cancel_api
    from nexusbridge.binance.spot.web_socket_stream import order_cancel_replace_api
    from nexusbridge.binance.spot.web_socket_stream import open_orders_status_api
    from nexusbridge.binance.spot.web_socket_stream import open_orders_cancel_all_api
    from nexusbridge.binance.spot.web_socket_stream import order_list_place_api
    from nexusbridge.binance.spot.web_socket_stream import order_place
    from nexusbridge.binance.spot.web_socket_stream import account_status_api
    from nexusbridge.binance.spot.web_socket_stream import rate_limit_order_api
    from nexusbridge.binance.spot.web_socket_stream import all_orders_api
    from nexusbridge.binance.spot.web_socket_stream import all_order_lists_api
    from nexusbridge.binance.spot.web_socket_stream import my_trades_api
    from nexusbridge.binance.spot.web_socket_stream import my_prevented_matches_api
    from nexusbridge.binance.spot.web_socket_stream import my_allocations_api
    from nexusbridge.binance.spot.web_socket_stream import account_commission_api



import hmac
import orjson
import asyncio

from typing import Any, Callable, Literal
from aiolimiter import AsyncLimiter

from nexusbridge.base import WSClient
from nexusbridge.bybit.constants import BybitAccountType


def generate_topic(
    base: str, symbol: str, depth: int = None, interval: str = None
) -> str:
    if base == "orderbook" and depth:
        return f"orderbook.{depth}.{symbol}"
    elif (
        base == "publicTrade"
        or base == "tickers"
        or base == "liquidation"
        or base == "tickers_lt"
        or base == "lt"
    ):
        return f"{base}.{symbol}"
    elif base == "kline" or base == "kline_lt":
        return f"{base}.{interval}.{symbol}"
    else:
        raise ValueError(f"Invalid base topic {base} or missing parameters")


class BybitWSClient(WSClient):
    def __init__(
        self,
        account_type: BybitAccountType,
        handler: Callable[..., Any],
        api_key: str = None,
        secret: str = None,
    ):
        self._account_type = account_type
        self._api_key = api_key
        self._secret = secret
        self._authed = False
        if self.is_private:
            url = account_type.ws_private_url
        else:
            url = account_type.ws_public_url
        # Bybit: do not exceed 500 requests per 5 minutes
        super().__init__(
            url,
            limiter=AsyncLimiter(max_rate=500, time_period=5 * 60),
            handler=handler,
            ping_idle_timeout=5,
            ping_reply_timeout=2,
            specific_ping_msg=orjson.dumps({"op": "ping"}),
            auto_ping_strategy="ping_when_idle",
        )

    @property
    def is_private(self):
        return self._api_key is not None or self._secret is not None

    def _generate_signature(self):
        expires = self._clock.timestamp_ms() + 1_000
        signature = str(
            hmac.new(
                bytes(self._secret, "utf-8"),
                bytes(f"GET/realtime{expires}", "utf-8"),
                digestmod="sha256",
            ).hexdigest()
        )
        return signature, expires

    def _get_auth_payload(self):
        signature, expires = self._generate_signature()
        return {"op": "auth", "args": [self._api_key, expires, signature]}

    async def _auth(self):
        if not self._authed:
            await self._send(self._get_auth_payload())
            self._authed = True
            await asyncio.sleep(5)

    async def _subscribe(self, topic: str | list[str], auth: bool = False):
        if isinstance(topic, list):
            for t in topic:
                if t in self._subscriptions:
                    self._log.debug(f"Already subscribed to {t}")
                    topic.remove(t)
                    return
        else:
            if topic in self._subscriptions:
                self._log.debug(f"Already subscribed to {topic}")
                return
        await self._connect()
        payload = {
            "op": "subscribe",
            "args": [topic] if isinstance(topic, str) else topic,
        }
        if auth:
            await self._auth()
        if isinstance(topic, list):
            for t in topic:
                self._subscriptions[t] = payload
        else:
            self._subscriptions[topic] = payload
        await self._send(payload)
        self._log.debug(f"Subscribing to {topic}.{self._account_type.value}...")

    async def subscribe_multiple(self, topics: list[str], auth: bool = False):
        await self._subscribe(topics, auth)

    async def subscribe_order_book(self, symbol: str, depth: int):
        """
        ### Linear & inverse:
        - Level 1 data, push frequency: 10ms
        - Level 50 data, push frequency: 20ms
        - Level 200 data, push frequency: 100ms
        - Level 500 data, push frequency: 100ms

        ### Spot:
        - Level 1 data, push frequency: 10ms
        - Level 50 data, push frequency: 20ms
        - Level 200 data, push frequency: 200ms

        ### Option:
        - Level 25 data, push frequency: 20ms
        - Level 100 data, push frequency: 100ms
        """
        topic = generate_topic("orderbook", symbol, depth)
        await self._subscribe(topic)

    async def subscribe_trade(self, symbol: str):
        topic = generate_topic("publicTrade", symbol)
        await self._subscribe(topic)

    async def subscribe_ticker(self, symbol: str):
        topic = generate_topic("tickers", symbol)
        await self._subscribe(topic)

    async def subscribe_kline(self, symbol: str, interval: str):
        """
        ### Available intervals:
        - 1 3 5 15 30 (min)
        - 60 120 240 360 720 (min)
        """
        topic = generate_topic("kline", symbol, interval=interval)
        await self._subscribe(topic)

    async def subscribe_liquidation(self, symbol: str):
        """
        Subscribe to the liquidation stream. Pushes at most one order per second per symbol. As such, this feed does not push all liquidations that occur on Bybit.
        """
        topic = generate_topic("liquidation", symbol)
        await self._subscribe(topic)

    async def subscribe_lt_kline(self, symbol: str, interval: str):
        """
        Available intervals:
        1 3 5 15 30 (min)
        60 120 240 360 720 (min)
        D (day)
        W (week)
        M (month)
        """
        topic = generate_topic("kline_lt", symbol, interval=interval)
        await self._subscribe(topic)

    async def subscribe_lt_ticker(self, symbol: str):
        topic = generate_topic("tickers_lt", symbol)
        await self._subscribe(topic)

    async def subscribe_lt_nav(self, symbol: str):
        topic = generate_topic("lt", symbol)
        await self._subscribe(topic)

    async def _resubscribe(self):
        if self.is_private:
            self._authed = False
            await self._auth()
        for _, payload in self._subscriptions.items():
            await self._send(payload)

    async def subscribe_order(
        self,
        topic: Literal[
            "order", "order.spot", "order.linear", "order.inverse", "order.option"
        ] = "order",
    ):
        """
        ### Topics:

        #### All in one:
        - order

        Categorical topics:
        - order.spot
        - order.linear
        - order.inverse
        - order.option
        """
        await self._subscribe(topic, auth=True)

    async def subscribe_position(
        self,
        topic: Literal[
            "position", "position.linear", "position.inverse", "position.option"
        ] = "position",
    ):
        """
        ### Topics:

        #### All in one:
        - position

        Categorical topics:
        - position.linear
        - position.inverse
        - position.option
        """
        await self._subscribe(topic, auth=True)

    async def subscribe_wallet(self, topic: str = "wallet"):
        """
        ### Topics:
        - wallet
        """
        await self._subscribe(topic, auth=True)

    async def subscribe_execution(
        self,
        topic: Literal[
            "execution",
            "execution.spot",
            "execution.linear",
            "execution.inverse",
            "execution.option",
        ] = "execution",
    ):
        """
        ### Topics:
        - execution
        """
        await self._subscribe(topic, auth=True)

    async def subscribe_fast_execution(
        self,
        topic: Literal[
            "execution.fast.linear",
            "execution.fast.inverse",
            "execution.fast.spot",
            "execution.fast",
        ] = "execution.fast",
    ):
        await self._subscribe(topic, auth=True)

    async def subscribe_greek(self, topic: str = "greeks"):
        """
        ### Topics:
        - greeks
        """
        await self._subscribe(topic, auth=True)

    async def subscribe_dcp(
        self, topic: Literal["dcp.future", "dcp.spot", "dcp.option"]
    ):
        """
        ### Topics:
        Subscribe to the dcp stream to trigger DCP function.
        For example, connection A subscribes "dcp.xxx", connection B does not and connection C subscribes "dcp.xxx".
            If A is alive, B is dead, C is alive, then this case will not trigger DCP.
            If A is alive, B is dead, C is dead, then this case will not trigger DCP.
            If A is dead, B is alive, C is dead, then DCP is triggered when reach the timeWindow threshold
            To sum up, for those private connections subscribing "dcp" topic are all dead, then DCP will be triggered.
        """
        await self._subscribe(topic, auth=True)



async def get_api_v3_account(self, **kwargs):
    """
    Get current account information.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/account-endpoints
    """
    endpoint = "/api/v3/account"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_api_v3_my_trades(self, symbol: str, **kwargs):
    """
    Get trades for a specific account and symbol.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/account-endpoints#account-trade-list-user_data
    """
    endpoint = "/api/v3/myTrades"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_api_v3_rate_limit_order(self, **kwargs):
    """
    Displays the user's unfilled order count for all intervals.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/account-endpoints#query-unfilled-order-count-user_data
    """
    endpoint = "/api/v3/rateLimit/order"
    payload = {**kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_api_v3_my_prevented_matches(self, symbol: str, **kwargs):
    """
    Displays the list of orders that were expired due to STP.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/account-endpoints#query-prevented-matches-user_data
    """
    endpoint = "/api/v3/myPreventedMatches"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_api_v3_my_allocations(self, symbol: str, **kwargs):
    """
    Retrieves allocations resulting from SOR order placement.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/account-endpoints#query-allocations-user_data
    """
    endpoint = "/api/v3/myAllocations"
    payload = {"symbol": symbol, **kwargs}
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw


async def get_api_v3_account_commission(self, symbol: str):
    """
    Get current account commission rates.
    https://developers.binance.com/docs/binance-spot-api-docs/rest-api/account-endpoints#query-commission-rates-user_data
    """
    endpoint = "/api/v3/account/commission"
    payload = {
        "symbol": symbol,
    }
    raw = await self._fetch("GET", endpoint, payload=payload, signed=True)
    return raw



async def get_fapi_v1_exchangeInfo(self):
    """
    Get the exchange information
    """
    endpoint = "/fapi/v1/exchangeInfo"
    raw = await self._fetch("GET", endpoint, signed=False)
    return raw



from typing import Literal, Callable, Any

from nexusbridge.binance.base import BinanceApiClient, BinanceWSClient
from nexusbridge.binance.constants import BinanceUrl, ALL_URL, BinanceInstrumentType


class PmTradingApi(BinanceApiClient):
    def __init__(self, key=None, secret=None, **kwargs):
        super().__init__(
            key=key,
            secret=secret,
            base_url=ALL_URL[BinanceInstrumentType.Derivatives_PM].get_url(
                BinanceUrl.MAIN
            ),
            **kwargs,
        )

    # Common endpoints
    from nexusbridge.binance.pm.common import get_papi_v1_ping
    from nexusbridge.binance.pm.common import post_papi_v1_listen_key
    from nexusbridge.binance.pm.common import put_papi_v1_listen_key
    from nexusbridge.binance.pm.common import delete_papi_v1_listen_key

    # Trade endpoints
    from nexusbridge.binance.pm.trade import post_papi_v1_um_order
    from nexusbridge.binance.pm.trade import post_papi_v1_um_conditional_order
    from nexusbridge.binance.pm.trade import post_papi_v1_cm_order
    from nexusbridge.binance.pm.trade import post_papi_v1_cm_conditional_order
    from nexusbridge.binance.pm.trade import post_papi_v1_margin_order
    from nexusbridge.binance.pm.trade import post_papi_v1_margin_loan
    from nexusbridge.binance.pm.trade import post_papi_v1_repay_loan
    from nexusbridge.binance.pm.trade import post_papi_v1_margin_order_oco
    from nexusbridge.binance.pm.trade import delete_papi_v1_um_order
    from nexusbridge.binance.pm.trade import delete_papi_v1_um_all_open_orders
    from nexusbridge.binance.pm.trade import delete_papi_v1_um_conditional_order
    from nexusbridge.binance.pm.trade import (
        delete_papi_v1_um_conditional_all_open_orders,
    )
    from nexusbridge.binance.pm.trade import delete_papi_v1_cm_order
    from nexusbridge.binance.pm.trade import delete_papi_v1_cm_all_open_orders
    from nexusbridge.binance.pm.trade import delete_papi_v1_cm_conditional_order
    from nexusbridge.binance.pm.trade import (
        delete_papi_v1_cm_conditional_all_open_orders,
    )
    from nexusbridge.binance.pm.trade import delete_papi_v1_margin_order
    from nexusbridge.binance.pm.trade import delete_papi_v1_margin_orderList
    from nexusbridge.binance.pm.trade import delete_papi_v1_margin_all_open_orders
    from nexusbridge.binance.pm.trade import put_papi_v1_um_order
    from nexusbridge.binance.pm.trade import put_papi_v1_cm_order
    from nexusbridge.binance.pm.trade import get_papi_v1_um_order
    from nexusbridge.binance.pm.trade import get_papi_v1_um_all_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_um_open_order
    from nexusbridge.binance.pm.trade import get_papi_v1_um_open_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_um_conditional_all_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_um_conditional_open_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_um_conditional_open_order
    from nexusbridge.binance.pm.trade import get_papi_v1_um_conditional_order_history
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_order
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_all_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_open_order
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_open_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_conditional_open_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_conditional_open_order
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_conditional_all_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_conditional_order_history
    from nexusbridge.binance.pm.trade import get_papi_v1_um_force_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_force_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_um_order_amendment
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_order_amendment
    from nexusbridge.binance.pm.trade import get_papi_v1_margin_force_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_um_user_trades
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_user_trades
    from nexusbridge.binance.pm.trade import get_papi_v1_um_adl_quantile
    from nexusbridge.binance.pm.trade import get_papi_v1_cm_adl_quantile
    from nexusbridge.binance.pm.trade import post_papi_v1_um_feeBurn
    from nexusbridge.binance.pm.trade import get_papi_v1_um_feeBurn
    from nexusbridge.binance.pm.trade import get_papi_v1_margin_order
    from nexusbridge.binance.pm.trade import get_papi_v1_margin_openOrders
    from nexusbridge.binance.pm.trade import get_papi_v1_margin_all_open_orders
    from nexusbridge.binance.pm.trade import get_papi_v1_margin_orderList
    from nexusbridge.binance.pm.trade import get_papi_v1_margin_all_orderList
    from nexusbridge.binance.pm.trade import get_papi_v1_margin_openOrderList
    from nexusbridge.binance.pm.trade import get_papi_v1_margin_myTrades
    from nexusbridge.binance.pm.trade import post_papi_v1_margin_repay_debt

    # Account endpoints
    from nexusbridge.binance.pm.account import get_papi_v1_balance
    from nexusbridge.binance.pm.account import get_papi_v1_account
    from nexusbridge.binance.pm.account import get_papi_v1_margin_max_borrowable
    from nexusbridge.binance.pm.account import get_papi_v1_margin_max_withdraw
    from nexusbridge.binance.pm.account import get_papi_v1_um_position_risk
    from nexusbridge.binance.pm.account import get_papi_v1_cm_position_risk
    from nexusbridge.binance.pm.account import post_papi_v1_um_leverage
    from nexusbridge.binance.pm.account import post_papi_v1_cm_leverage
    from nexusbridge.binance.pm.account import post_papi_v1_um_position_side_dual
    from nexusbridge.binance.pm.account import post_papi_v1_cm_position_side_dual
    from nexusbridge.binance.pm.account import get_papi_v1_um_position_side_dual
    from nexusbridge.binance.pm.account import get_papi_v1_cm_position_side_dual
    from nexusbridge.binance.pm.account import get_papi_v1_um_leverage_bracket
    from nexusbridge.binance.pm.account import get_papi_v1_cm_leverage_bracket
    from nexusbridge.binance.pm.account import get_papi_v1_um_api_trading_status
    from nexusbridge.binance.pm.account import get_papi_v1_um_commission_rate
    from nexusbridge.binance.pm.account import get_papi_v1_cm_commission_rate
    from nexusbridge.binance.pm.account import get_papi_v1_margin_margin_loan
    from nexusbridge.binance.pm.account import get_papi_v1_margin_repay_loan
    from nexusbridge.binance.pm.account import get_papi_v1_repay_futures_switch
    from nexusbridge.binance.pm.account import post_papi_v1_repay_futures_switch
    from nexusbridge.binance.pm.account import (
        get_papi_v1_margin_margin_interest_history,
    )
    from nexusbridge.binance.pm.account import (
        post_papi_v1_repay_futures_negative_balance,
    )
    from nexusbridge.binance.pm.account import get_papi_v1_portfolio_interest_history
    from nexusbridge.binance.pm.account import post_papi_v1_auto_collection
    from nexusbridge.binance.pm.account import post_papi_v1_asset_collection
    from nexusbridge.binance.pm.account import post_papi_v1_bnb_transfer
    from nexusbridge.binance.pm.account import get_papi_v1_um_income
    from nexusbridge.binance.pm.account import get_papi_v1_cm_income
    from nexusbridge.binance.pm.account import get_papi_v1_um_account
    from nexusbridge.binance.pm.account import get_papi_v1_cm_account
    from nexusbridge.binance.pm.account import get_papi_v1_um_account_config
    from nexusbridge.binance.pm.account import get_papi_v1_um_symbol_config
    from nexusbridge.binance.pm.account import get_papi_v2_um_account
    from nexusbridge.binance.pm.account import get_papi_v1_um_trade_asyn
    from nexusbridge.binance.pm.account import get_papi_v1_um_trade_asyn_id
    from nexusbridge.binance.pm.account import get_papi_v1_um_order_asyn
    from nexusbridge.binance.pm.account import get_papi_v1_um_order_asyn_id
    from nexusbridge.binance.pm.account import get_papi_v1_um_income_asyn
    from nexusbridge.binance.pm.account import get_papi_v1_um_income_asyn_id
    from nexusbridge.binance.pm.account import get_papi_v1_rate_limit_order


class PmTradingWebsocket(BinanceWSClient):
    def __init__(
        self,
        handler: Callable[..., Any],
        key=None,
        secret=None,
        binance_url: Literal[BinanceUrl.WS, BinanceUrl.TEST_WS] = BinanceUrl.WS,
        **kwargs,
    ):
        if "base_url" not in kwargs:
            self.baseUrl = ALL_URL[BinanceInstrumentType.SPOT].get_url(binance_url)
        else:
            self.baseUrl = kwargs["base_url"]
        super().__init__(
            url=self.baseUrl, handler=handler, key=key, secret=secret, **kwargs
        )
